#!/usr/bin/env python

# Copyright, license and disclaimer are at the end of this file.

# This is the latest, enhanced version of the asizeof.py recipes at
# <http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/546530>
# <http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/544288>

'''This module exposes 10 functions and 2 classes to obtain lengths
   and sizes of Python objects (for Python 2.2 or later [1]).

   The main changes in this version are new function  calcsize(),
   use  gc.get_objects() to get all objects and improvements in
   this documentation.

Public Functions [2]

   Function  asizeof calculates the combined (approximate) size
   in bytes of one or several Python objects.

   Function  asizesof returns a tuple containing the (approximate)
   size in bytes for each given Python object separately.

   Function  asized returns for each object an instance of class
   Asized containing all the size information of the object and a
   tuple with the referents.

   Functions  basicsize and  itemsize return the basic respectively
   item size of the given object.

   Function  flatsize returns the flat size of a Python object in
   bytes defined as the basic size plus the item size times the
   length of the given object.

   Function  leng returns the length of an object, like standard
   len but extended for several types, e.g. the  leng of a multi-
   precision int (or long) is the number of digits [3].  The length
   of most mutable sequence objects includes an estimate of the
   over-allocation and therefore, the  leng value may differ from
   the standard  len result.

   Function  refs returns (a generator for) the referents of the
   given object, i.e. the objects referenced by the given object.

   Function  calcsize is equivalent to standard  struct.calcsize
   but handles format characters 'z' for signed C type  Py_ssize_t
   and 'Z' for unsigned C type  size_t.

   Certain classes are known to be sub-classes of or to behave as
   dict objects.  Function  adict can be used to install other
   class objects to be treated like dict.

Public Classes [2]

   An instance of class  Asized is returned for each object sized
   with the  asized function or method.

   Class  Asizer can be used to accumulate the results of several
   asizeof or  asizesof calls.  After creating an  Asizer instance,
   use methods  asizeof and  asizesof to size additional objects.

   Call methods  exclude_refs and/or  exclude_types to exclude
   references to or instances or types of certain objects.

   Use one of the  print\_... methods to report the statistics.

Duplicate Objects

   Any duplicate, given objects are sized only once and the size
   is included in the combined total only once.  But functions
   asizesof and  asized do return a size value respectively an
   Asized instance for each given object, the same for duplicates.

Definitions [4]

   The size of an object is defined as the sum of the flat size
   of the object plus the sizes of any referents.  Referents are
   visited recursively up to a given limit.  However, the size
   of objects referenced multiple times is included only once.

   The flat size of an object is defined as the basic size of the
   object plus the item size times the number of allocated items.
   The flat size does include the size for the items (references
   to the referents), but not the referents themselves.

   The flat size returned by function  flatsize equals the result
   of the  asizeof function with options code=True, ignored=False,
   limit=0 and option  align set to the same value.

   The accurate flat size for an object is obtained from function
   sys.getsizeof() where available.  Otherwise, the length and
   size of sequence objects as dicts, lists, sets, etc. is based
   on an estimate for the number of allocated items.  As a result,
   the reported length and size may substantially differ from the
   actual length and size.

   The basic and item sizes are obtained from the __basicsize__
   respectively __itemsize__ attribute of the (type of the) object.
   Where necessary (e.g. sequence objects), a zero __itemsize__
   is replaced by the size of a corresponding C type.

   The basic size (of GC managed objects) objects includes the
   overhead for Python's garbage collector (GC) as well as the
   space needed for refcounts (only in certain Python builds).

   Optionally, sizes can be aligned to any power of 2 multiple.

Size of (byte)code

   The (byte)code size of objects as classes, functions, methods,
   modules, etc. can be included by setting option  code.

   Iterators are handled similar to sequences: iterated object(s)
   are sized like referents if the recursion limit permits.  Also,
   function  gc.get_referents() must return the referent object
   of iterators.

   Generators are sized as (byte)code only, but generated objects
   are never sized.

Old- and New-style Classes

   All old- and new-style class, instance and type objects, are
   handled uniformly such that (a) instance and class objects can
   be distinguished and (b) instances of different old-style
   classes can be dealt with separately.

   Class and type objects are represented as <class ....* def>
   respectively as <type ... def> where an '*' indicates an old-
   style class and the  def suffix marks the definition object.
   Instances of old-style classes are shown as new-style ones but
   with an '*' at the end of the name, like <class module.name*>.

Ignored Objects

   To avoid excessive sizes, several object types are ignored [4]
   by default, e.g. built-in functions, built-in types and classes
   [5], function globals and module referents.  However, any
   instances thereof are sized and module objects will be sized
   when passed as given objects.  Ignored object types are included
   if option  ignored is set accordingly.

   In addition, many __...__ attributes of callable objects are
   ignored, except crucial ones, e.g. class attributes __dict__,
   __doc__, __name__ and __slots__.  For more details, see the
   type-specific _..._refs() and _len_...() functions below.

   Option  all can be used to size all Python objects and/or get
   the referents from  gc.get_referents() and override the type-
   specific __..._refs() functions.

Notes

   [1] Tested with Python 2.2.3, 2.3.7, 2.4.5, 2.5.1, 2.5.2, 2.6.2,
       3.0.1 or 3.1a2 on CentOS 4.6, SuSE 9.3, MacOS X 10.4.11 Tiger
       (Intel) and 10.3.9 Panther (PPC), Solaris 10 (Opteron) and
       Windows XP all 32-bit Python and on RHEL 3u7 and Solaris 10
       (Opteron) both 64-bit Python.

   [2] The functions and classes in this module are not thread-safe.

   [3] See Python source file .../Include/longinterp.h for the
       C typedef of digit used in multi-precision int (or long)
       objects.  The size of digit in bytes can be obtained in
       Python from the int (or long) __itemsize__ attribute.
       Function  leng (rather _len_int) below deterimines the
       number of digits from the int (or long) value.

   [4] These definitions and other assumptions are rather arbitrary
       and may need corrections or adjustments.

   [5] Types and classes are considered built-in if the module of
       the type or class is listed in  _builtin_modules below.

'''  #PYCHOK expected

from __future__ import generators  #PYCHOK for yield in Python 2.2

from inspect    import isbuiltin, isclass, iscode, isframe, \
                       isfunction, ismethod, ismodule, stack
from math       import log
from os         import linesep
from struct     import calcsize as _calcsize
import sys
import types    as     Types
import weakref  as     Weakref

__version__ = '5.12 (Apr 27, 2009)'
__all__     = ['adict', 'asized', 'asizeof', 'asizesof',
               'Asized', 'Asizer',  # classes
               'basicsize', 'flatsize', 'itemsize', 'leng', 'refs',
               'calcsize']  # handles 'z' and 'Z'

 # any classes or types in modules listed in _builtin_modules are
 # considered built-in and ignored by default, as built-in functions
if __name__ == '__main__':
    _builtin_modules = (int.__module__, 'types', Exception.__module__)  # , 'weakref'
else:  # treat this very module as built-in
    _builtin_modules = (int.__module__, 'types', Exception.__module__, __name__)  # , 'weakref'

 # sizes of some primitive C types
 # XXX len(pack(T, 0)) == Struct(T).size == _calcsize(T)
 # but type/class Struct only available since Python 2.5
_sizeof_Cbyte  = _calcsize('c')  # sizeof(unsigned char)
_sizeof_Clong  = _calcsize('l')  # sizeof(long)
_sizeof_Cvoidp = _calcsize('P')  # sizeof(void*)

 # sizeof(long) != sizeof(ssize_t) on LLP64
if _sizeof_Clong < _sizeof_Cvoidp:
    _Zz = 'PP'
else:
    _Zz = 'Ll'

def calcsize(fmt):
    '''struct.calcsize() handling 'z' for signed Py_ssize_t and 'Z' for unsigned size_t.
    '''
    return _calcsize(fmt.replace('Z', _Zz[0]).replace('z', _Zz[1]))

 # defaults for some basic sizes with 'z' for C Py_ssize_t
_sizeof_CPyCodeObject   = calcsize('Pz10P5i0P')    # sizeof(PyCodeObject)
_sizeof_CPyFrameObject  = calcsize('Pzz13P63i0P')  # sizeof(PyFrameObject)
_sizeof_CPyModuleObject = calcsize('PzP0P')        # sizeof(PyModuleObject)

 # defaults for some item sizes with 'z' for C Py_ssize_t
_sizeof_CPyDictEntry    = calcsize('z2P')  # sizeof(PyDictEntry)
_sizeof_Csetentry       = calcsize('lP')   # sizeof(setentry)

 # XXX use sys.int_info.sizeof_digit in Python 3.1
try:  # C typedef digit for multi-precision int (or long)
    _sizeof_Cdigit = long.__itemsize__
except NameError:  # no long in Python 3.0
    _sizeof_Cdigit = int.__itemsize__
if _sizeof_Cdigit < 2:
    raise AssertionError('sizeof(%s) bad: %d' % ('digit', _sizeof_Cdigit))

try:  # sizeof(unicode_char)
    u = unicode('\0')
except NameError:  # no unicode() in Python 3.0
    u = '\0'
u = u.encode('unicode-internal')  # see .../Lib/test/test_sys.py
_sizeof_Cunicode = len(u)
del u
if (1 << (_sizeof_Cunicode << 3)) <= sys.maxunicode:
    raise AssertionError('sizeof(%s) bad: %d' % ('unicode', _sizeof_Cunicode))

if hasattr(sys, 'maxsize'):  # new in Python 2.6
    Z = calcsize('Z')  # check sizeof(size_t)
    if (1 << (Z << 3)) <= sys.maxsize:
        raise AssertionError('sizeof(%s) bad: %d' % ('size_t', Z))
    del Z

try:  # size of GC header, sizeof(PyGC_Head)
    import _testcapi as t
    _sizeof_CPyGC_Head = t.SIZEOF_PYGC_HEAD  # new in Python 2.6
except (ImportError, AttributeError):  # sizeof(PyGC_Head)
     # alignment should be to sizeof(long double) but there
     # is no way to obtain that value, assume twice double
    t = calcsize('2d') - 1
    _sizeof_CPyGC_Head = (calcsize('2Pz') + t) & ~t
del t

 # size of refcounts (Python debug build only)
if hasattr(sys, 'gettotalrefcount'):
    _sizeof_Crefcounts = calcsize('2z')
else:
    _sizeof_Crefcounts = 0

 # some flags from .../Include/object.h
_Py_TPFLAGS_HEAPTYPE = 1 <<  9  # Py_TPFLAGS_HEAPTYPE
_Py_TPFLAGS_HAVE_GC  = 1 << 14  # Py_TPFLAGS_HAVE_GC

_Type_type = type(type)  # == type and new-style class type


 # compatibility functions for more uniform
 # behavior across Python version 2.2 thu 3.0

def _items(obj):  # dict only
    '''Return iter-/generator, preferably.
    '''
    return getattr(obj, 'iteritems', obj.items)()

def _keys(obj):  # dict only
    '''Return iter-/generator, preferably.
    '''
    return getattr(obj, 'iterkeys', obj.keys)()

def _values(obj):  # dict only
    '''Use iter-/generator, preferably.
    '''
    return getattr(obj, 'itervalues', obj.values)()

try:  # callable() builtin
    _callable = callable
except NameError:  # callable() removed in Python 3.0
    def _callable(obj):
        '''Substitute for callable().'''
        return hasattr(obj, '__call__')

try:  # get 'all' current objects
    from gc import get_objects as _getobjects
except ImportError:
    def _getobjects():  # pyflakes:ignore
         # modules first, globals and stack
         # (may contain duplicate objects)
        return tuple(_values(sys.modules)) + (
               globals(), stack(sys.getrecursionlimit()))

try:  # get 'all' referents of objects
      # note that gc.get_referents()
      # returns () for dict...-iterators
    from gc import get_referents as _getreferents
except ImportError:  # no get_referents() in Python 2.2
    def _getreferents(unused):  # pyflakes:ignore
        return ()  # sorry, no refs

 # sys.getsizeof() new in Python 2.6
_getsizeof = getattr(sys, 'getsizeof', None)

try:  # str intern()
    _intern = intern
except NameError:  # no intern() in Python 3.0
    def _intern(val):
        return val

def _kwds(**kwds):  # no dict(key=value, ...) in Python 2.2
    '''Return name=value pairs as keywords dict.
    '''
    return kwds

try:  # sorted() builtin
    _sorted = sorted
except NameError:  # no sorted() in Python 2.2
    def _sorted(vals, reverse=False):
        '''Partial substitute for missing sorted().'''
        vals.sort()  # inplace OK
        if reverse:
            vals.reverse()
        return vals

try:  # sum() builtin
    _sum = sum
except NameError:  # no sum() in Python 2.2
    def _sum(vals):
        '''Partial substitute for missing sum().'''
        s = 0
        for v in vals:
            s += v
        return s


 # private functions

def _basicsize(t, base=0, heap=False, obj=None):
    '''Get non-zero basicsize of type,
       including the header sizes.
    '''
    s = max(getattr(t, '__basicsize__', 0), base)
     # include gc header size
    if t != _Type_type:
       h = getattr(t,   '__flags__', 0) & _Py_TPFLAGS_HAVE_GC
    elif heap:  # type, allocated on heap
       h = True
    else:  # None has no __flags__ attr
       h = getattr(obj, '__flags__', 0) & _Py_TPFLAGS_HEAPTYPE
    if h:
       s += _sizeof_CPyGC_Head
     # include reference counters
    return s + _sizeof_Crefcounts

def _derive_typedef(typ):
    '''Return single, existing super type typedef or None.
    '''
    v = [v for v in _values(_typedefs) if _issubclass(typ, v.type)]
    if len(v) == 1:
        return v[0]
    return None

def _dir2(obj, pref='', excl=(), slots=None, itor=''):
    '''Return an attribute name, object 2-tuple for certain
       attributes or for the '__slots__' attributes of the
       given object, but not both.  Any iterator referent
       objects are returned with the given name if the
       latter is non-empty.
    '''
    if slots:  # __slots__ attrs
        if hasattr(obj, slots):
             # collect all inherited __slots__ attrs
             # from list, tuple, or dict __slots__,
             # while removing any duplicate attrs
            s = {}
            for c in type(obj).mro():
                for a in getattr(c, slots, ()):
                    if hasattr(obj, a):
                        s.setdefault(a, getattr(obj, a))
             # assume __slots__ tuple/list
             # is holding the attr values
            yield slots, _Slots(s)  # _keys(s)
            for t in _items(s):
                yield t  # attr name, value
    elif itor:  # iterator referents
        for o in obj:  # iter(obj)
            yield itor, o
    else:  # regular attrs
        for a in dir(obj):
            if a.startswith(pref) and a not in excl and hasattr(obj, a):
               yield a, getattr(obj, a)

def _infer_dict(obj):
    '''Return True for likely dict object.
    '''
    for ats in (('__len__', 'get', 'has_key',     'items',     'keys',     'values'),
                ('__len__', 'get', 'has_key', 'iteritems', 'iterkeys', 'itervalues')):
        for a in ats:  # no all(<generator_expression>) in Python 2.2
            if not _callable(getattr(obj, a, None)):
                break
        else:  # all True
            return True
    return False

def _isdictclass(obj):
    '''Return True for known dict objects.
    '''
    c = getattr(obj, '__class__', None)
    return c and c.__name__ in _dict_classes.get(c.__module__, ())

def _issubclass(sub, sup):
    '''Safe issubclass().
    '''
    if sup is not object:
        try:
            return issubclass(sub, sup)
        except TypeError:
            pass
    return False

def _itemsize(t, item=0):
    '''Get non-zero itemsize of type.
    '''
     # replace zero value with default
    return getattr(t, '__itemsize__', 0) or item

def _kwdstr(**kwds):
    '''Keyword arguments as a string.
    '''
    return ', '.join(_sorted(['%s=%r' % kv for kv in _items(kwds)]))  # [] for Python 2.2

def _lengstr(obj):
    '''Object length as a string.
    '''
    n = leng(obj)
    if n is None:  # no len
       r = ''
    elif n > _len(obj):  # extended
       r = ' leng %d!' % n
    else:
       r = ' leng %d' % n
    return r

def _nameof(obj, dflt=''):
    '''Return the name of an object.
    '''
    return getattr(obj, '__name__', dflt)

def _objs(objs, all=None, **unused):
    '''Return the given or 'all' objects.
    '''
    if all in (False, None):
        t = objs or ()
    elif all is True:  # 'all' objects
        t = objs or _getobjects()
    else:
        raise ValueError('invalid option: %s=%r' % ('all', all))
    return t

def _p100(part, total, prec=1):
    '''Return percentage as string.
    '''
    r = float(total)
    if r:
        r = part * 100.0 / r
        return '%.*f%%' % (prec, r)
    return 'n/a'

def _plural(num):
    '''Return 's' if plural.
    '''
    if num == 1:
        s = ''
    else:
        s = 's'
    return s

def _power2(n):
    '''Find the next power of 2.
    '''
    p2 = 16
    while n > p2:
        p2 += p2
    return p2

def _prepr(obj, clip=0):
    '''Prettify and clip long repr() string.
    '''
    return _repr(obj, clip=clip).strip('<>').replace("'", '')  # remove <''>

def _printf(fmt, *args, **print3opts):
    '''Formatted print.
    '''
    if print3opts:  # like Python 3.0
        f = print3opts.get('file', None) or sys.stdout
        if args:
            f.write(fmt % args)
        else:
            f.write(fmt)
        f.write(print3opts.get('end', linesep))
    elif args:
        print(fmt % args)
    else:
        print(fmt)

def _refs(obj, named, *ats, **kwds):
    '''Return specific attribute objects of an object.
    '''
    if named:
        for a in ats:  # cf. inspect.getmembers()
            if hasattr(obj, a):
                yield _NamedRef(a, getattr(obj, a))
        if kwds:  # kwds are _dir2() args
            for a, o in _dir2(obj, **kwds):
                yield _NamedRef(a, o)
    else:
        for a in ats:  # cf. inspect.getmembers()
            if hasattr(obj, a):
                yield getattr(obj, a)
        if kwds:  # kwds are _dir2() args
            for _, o in _dir2(obj, **kwds):
                yield o

def _repr(obj, clip=80):
    '''Clip long repr() string.
    '''
    try:  # safe repr()
        r = repr(obj)
    except TypeError:
        r = 'N/A'
    if 0 < clip < len(r):
        h = (clip // 2) - 2
        if h > 0:
            r = r[:h] + '....' + r[-h:]
    return r

def _SI(size, K=1024, i='i'):
    '''Return size as SI string.
    '''
    if 1 < K < size:
        f = float(size)
        for si in iter('KMGPTE'):
            f /= K
            if f < K:
                return ' or %.1f %s%sB' % (f, si, i)
    return ''

def _SI2(size, **kwds):
    '''Return size as regular plus SI string.
    '''
    return str(size) + _SI(size, **kwds)

 # type-specific referent functions

def _class_refs(obj, named):
    '''Return specific referents of a class object.
    '''
    return _refs(obj, named, '__class__', '__dict__',  '__doc__', '__mro__',
                             '__name__',  '__slots__', '__weakref__')

def _co_refs(obj, named):
    '''Return specific referents of a code object.
    '''
    return _refs(obj, named, pref='co_')

def _dict_refs(obj, named):
    '''Return key and value objects of a dict/proxy.
    '''
    if named:
        for k, v in _items(obj):
            s = str(k)
            yield _NamedRef(s, k, 1)  # key
            yield _NamedRef(s, v, 2)  # value
    else:
        for k, v in _items(obj):
            yield k
            yield v

def _enum_refs(obj, named):
    '''Return specific referents of an enumerate object.
    '''
    return _refs(obj, named, '__doc__')

def _exc_refs(obj, named):
    '''Return specific referents of an Exception object.
    '''
     # .message raises DeprecationWarning in Python 2.6
    return _refs(obj, named, 'args', 'filename', 'lineno', 'msg', 'text')  # , 'message', 'mixed'

def _file_refs(obj, named):
    '''Return specific referents of a file object.
    '''
    return _refs(obj, named, 'mode', 'name')

def _frame_refs(obj, named):
    '''Return specific referents of a frame object.
    '''
    return _refs(obj, named, pref='f_')

def _func_refs(obj, named):
    '''Return specific referents of a function or lambda object.
    '''
    return _refs(obj, named, '__doc__', '__name__', '__code__',
                             pref='func_', excl=('func_globals',))

def _gen_refs(obj, named):
    '''Return the referent(s) of a generator object.
    '''
     # only some gi_frame attrs
    f = getattr(obj, 'gi_frame', None)
    return _refs(f, named, 'f_locals', 'f_code')

def _im_refs(obj, named):
    '''Return specific referents of a method object.
    '''
    return _refs(obj, named, '__doc__', '__name__', '__code__',
                             pref='im_')

def _inst_refs(obj, named):
    '''Return specific referents of a class instance.
    '''
    return _refs(obj, named, '__dict__', '__class__',
                             slots='__slots__')

def _iter_refs(obj, named):
    '''Return the referent(s) of an iterator object.
    '''
    r = _getreferents(obj)  # special case
    return _refs(r, named, itor=_nameof(obj) or 'iteref')

def _module_refs(obj, named):
    '''Return specific referents of a module object.
    '''
     # ignore this very module
    if obj.__name__ == __name__:
        return ()
     # module is essentially a dict
    return _dict_refs(obj.__dict__, named)

def _prop_refs(obj, named):
    '''Return specific referents of a property object.
    '''
    return _refs(obj, named, '__doc__', pref='f')

def _seq_refs(obj, unused):  # named unused for PyChecker
    '''Return specific referents of a frozen/set, list, tuple and xrange object.
    '''
    return obj  # XXX for r in obj: yield r

def _stat_refs(obj, named):
    '''Return referents of a os.stat object.
    '''
    return _refs(obj, named, pref='st_')

def _statvfs_refs(obj, named):
    '''Return referents of a os.statvfs object.
    '''
    return _refs(obj, named, pref='f_')

def _tb_refs(obj, named):
    '''Return specific referents of a traceback object.
    '''
    return _refs(obj, named, pref='tb_')

def _type_refs(obj, named):
    '''Return specific referents of a type object.
    '''
    return _refs(obj, named, '__dict__', '__doc__', '__mro__',
                             '__name__', '__slots__', '__weakref__')

def _weak_refs(obj, unused):  # named unused for PyChecker
    '''Return weakly referent object.
    '''
    try:  # ignore 'key' of KeyedRef
        return (obj(),)
    except:  # XXX ReferenceError
        return ()  #PYCHOK OK

_all_refs = (None, _class_refs,   _co_refs,   _dict_refs,  _enum_refs,
                   _exc_refs,     _file_refs, _frame_refs, _func_refs,
                   _gen_refs,     _im_refs,   _inst_refs,  _iter_refs,
                   _module_refs,  _prop_refs, _seq_refs,   _stat_refs,
                   _statvfs_refs, _tb_refs,   _type_refs,  _weak_refs)


 # type-specific length functions

def _len(obj):
    '''Safe len().
    '''
    try:
        return len(obj)
    except TypeError:  # no len()
        return 0

def _len_array(obj):
    '''Array length in bytes.
    '''
    return len(obj) * obj.itemsize

def _len_bytearray(obj):
    '''Bytearray size.
    '''
    return obj.__alloc__()

def _len_code(obj):  # see .../Lib/test/test_sys.py
    '''Length of code object (stack and variables only).
    '''
    return obj.co_stacksize +      obj.co_nlocals  \
                            + _len(obj.co_freevars) \
                            + _len(obj.co_cellvars) - 1

def _len_dict(obj):
    '''Dict length in items (estimate).
    '''
    n = len(obj)  # active items
    if n < 6:  # ma_smalltable ...
       n = 0  # ... in basicsize
    else:  # at least one unused
       n = _power2(n + 1)
    return n

def _len_frame(obj):
    '''Length of a frame object.
    '''
    c = getattr(obj, 'f_code', None)
    if c:
       n = _len_code(c)
    else:
       n = 0
    return n

_digit2p2 =  1 << (_sizeof_Cdigit << 3)
_digitmax = _digit2p2 - 1  # == (2 * PyLong_MASK + 1)
_digitlog =  1.0 / log(_digit2p2)

def _len_int(obj):
    '''Length of multi-precision int (aka long) in digits.
    '''
    if obj:
        n, i = 1, abs(obj)
        if i > _digitmax:
             # no log(x[, base]) in Python 2.2
            n += int(log(i) * _digitlog)
    else:  # zero
        n = 0
    return n

def _len_iter(obj):
    '''Length (hint) of an iterator.
    '''
    n = getattr(obj, '__length_hint__', None)
    if n:
       n = n()
    else:  # try len()
       n = _len(obj)
    return n

def _len_list(obj):
    '''Length of list (estimate).
    '''
    n = len(obj)
     # estimate over-allocation
    if n > 8:
       n += 6 + (n >> 3)
    elif n:
       n += 4
    return n

def _len_module(obj):
    '''Module length.
    '''
    return _len(obj.__dict__)  # _len(dir(obj))

def _len_set(obj):
    '''Length of frozen/set (estimate).
    '''
    n = len(obj)
    if n > 8:  # assume half filled
       n = _power2(n + n - 2)
    elif n:  # at least 8
       n = 8
    return n

def _len_slice(obj):
    '''Slice length.
    '''
    try:
        return ((obj.stop - obj.start + 1) // obj.step)
    except (AttributeError, TypeError):
        return 0

def _len_slots(obj):
    '''Slots length.
    '''
    return len(obj) - 1

def _len_struct(obj):
    '''Struct length in bytes.
    '''
    try:
        return obj.size
    except AttributeError:
        return 0

def _len_unicode(obj):
    '''Unicode size.
    '''
    return len(obj) + 1

_all_lengs = (None, _len,        _len_array,  _len_bytearray,
                    _len_code,   _len_dict,   _len_frame,
                    _len_int,    _len_iter,   _len_list,
                    _len_module, _len_set,    _len_slice,
                    _len_slots,  _len_struct, _len_unicode)


# more private functions and classes

_old_style = '*'  # marker
_new_style = ''   # no marker

class _Claskey(object):
    '''Wrapper for class objects.
    '''
    __slots__ = ('_obj', '_sty')

    def __init__(self, obj, style):
        self._obj = obj  # XXX Weakref.ref(obj)
        self._sty = style

    def __str__(self):
        r = str(self._obj)
        if r.endswith('>'):
            r = '%s%s def>' % (r[:-1], self._sty)
        elif self._sty is _old_style and not r.startswith('class '):
            r = 'class %s%s def' % (r, self._sty)
        else:
            r = '%s%s def' % (r, self._sty)
        return r
    __repr__ = __str__

 # For most objects, the object type is used as the key in the
 # _typedefs dict further below, except class and type objects
 # and old-style instances.  Those are wrapped with separate
 # _Claskey or _Instkey instances to be able (1) to distinguish
 # instances of different old-style classes by class, (2) to
 # distinguish class (and type) instances from class (and type)
 # definitions for new-style classes and (3) provide similar
 # results for repr() and str() of new- and old-style classes
 # and instances.

_claskeys = {}  # [id(obj)] = _Claskey()

def _claskey(obj, style):
    '''Wrap an old- or new-style class object.
    '''
    i =  id(obj)
    k = _claskeys.get(i, None)
    if not k:
        _claskeys[i] = k = _Claskey(obj, style)
    return k

try:  # no Class- and InstanceType in Python 3.0
    _Types_ClassType    = Types.ClassType
    _Types_InstanceType = Types.InstanceType

    class _Instkey(object):
        '''Wrapper for old-style class (instances).
        '''
        __slots__ = ('_obj',)

        def __init__(self, obj):
            self._obj = obj  # XXX Weakref.ref(obj)

        def __str__(self):
            return '<class %s.%s%s>' % (self._obj.__module__, self._obj.__name__, _old_style)
        __repr__ = __str__

    _instkeys = {}  # [id(obj)] = _Instkey()

    def _instkey(obj):
        '''Wrap an old-style class (instance).
        '''
        i =  id(obj)
        k = _instkeys.get(i, None)
        if not k:
            _instkeys[i] = k = _Instkey(obj)
        return k

    def _keytuple(obj):
        '''Return class and instance keys for a class.
        '''
        t = type(obj)
        if t is _Types_InstanceType:
            t = obj.__class__
            return _claskey(t,   _old_style), _instkey(t)
        elif t is _Types_ClassType:
            return _claskey(obj, _old_style), _instkey(obj)
        elif t is _Type_type:
            return _claskey(obj, _new_style), obj
        return None, None  # not a class

    def _objkey(obj):
        '''Return the key for any object.
        '''
        k = type(obj)
        if k is _Types_InstanceType:
            k = _instkey(obj.__class__)
        elif k is _Types_ClassType:
            k = _claskey(obj, _old_style)
        elif k is _Type_type:
            k = _claskey(obj, _new_style)
        return k

except AttributeError:  # Python 3.0

    def _keytuple(obj):  #PYCHOK expected  # pyflakes:ignore
        '''Return class and instance keys for a class.
        '''
        if type(obj) is _Type_type:  # isclass(obj):
            return _claskey(obj, _new_style), obj
        return None, None  # not a class

    def _objkey(obj):  #PYCHOK expected  # pyflakes:ignore
        '''Return the key for any object.
        '''
        k = type(obj)
        if k is _Type_type:  # isclass(obj):
            k = _claskey(obj, _new_style)
        return k

class _NamedRef(object):
    '''Store referred object along
       with the name of the referent.
    '''
    __slots__ = ('name', 'ref', 'typ')

    def __init__(self, name, ref, typ=0):
        self.name = name
        self.ref  = ref
        self.typ  = typ  # see Asized.format

class _Slots(tuple):
    '''Wrapper class for __slots__ attribute at
       class instances to account for the size
       of the __slots__ tuple/list containing
       references to the attribute values.
    '''
    pass

 # kinds of _Typedefs
_i = _intern
_all_kinds = (_kind_static, _kind_dynamic, _kind_derived, _kind_ignored, _kind_inferred) = (
                _i('static'), _i('dynamic'), _i('derived'), _i('ignored'), _i('inferred'))
del _i

class _Typedef(object):
    '''Type definition class.
    '''
    __slots__ = {
        'base': 0,     # basic size in bytes
        'item': 0,     # item size in bytes
        'leng': None,  # or _len_...() function
        'refs': None,  # or _..._refs() function
        'both': None,  # both data and code if True, code only if False
        'kind': None,  # _kind_... value
        'type': None}  # original type

    def __init__(self, **kwds):
        self.reset(**kwds)

    def __lt__(self, unused):  # for Python 3.0
        return True

    def __repr__(self):
        return repr(self.args())

    def __str__(self):
        t = [str(self.base), str(self.item)]
        for f in (self.leng, self.refs):
            if f:
                t.append(f.__name__)
            else:
                t.append('n/a')
        if not self.both:
            t.append('(code only)')
        return ', '.join(t)

    def args(self):  # as args tuple
        '''Return all attributes as arguments tuple.
        '''
        return (self.base, self.item, self.leng, self.refs,
                self.both, self.kind, self.type)

    def dup(self, other=None, **kwds):
        '''Duplicate attributes of dict or other typedef.
        '''
        if other is None:
            d = _dict_typedef.kwds()
        else:
            d =  other.kwds()
        d.update(kwds)
        self.reset(**d)

    def flat(self, obj, mask=0):
        '''Return the aligned flat size.
        '''
        s = self.base
        if self.leng and self.item > 0:  # include items
            s += self.leng(obj) * self.item
        if _getsizeof:  # _getsizeof prevails
            s = _getsizeof(obj, s)
        if mask:  # align
            s = (s + mask) & ~mask
        return s

    def format(self):
        '''Return format dict.
        '''
        c = n = ''
        if not self.both:
            c = ' (code only)'
        if self.leng:
            n = ' (%s)' % _nameof(self.leng)
        return _kwds(base=self.base, item=self.item, leng=n,
                     code=c,         kind=self.kind)

    def kwds(self):
        '''Return all attributes as keywords dict.
        '''
         # no dict(refs=self.refs, ..., kind=self.kind) in Python 2.0
        return _kwds(base=self.base, item=self.item,
                     leng=self.leng, refs=self.refs,
                     both=self.both, kind=self.kind, type=self.type)

    def save(self, t, base=0, heap=False):
        '''Save this typedef plus its class typedef.
        '''
        c, k = _keytuple(t)
        if k and k not in _typedefs:  # instance key
            _typedefs[k] = self
            if c and c not in _typedefs:  # class key
                if t.__module__ in _builtin_modules:
                    k = _kind_ignored  # default
                else:
                    k = self.kind
                _typedefs[c] = _Typedef(base=_basicsize(type(t), base=base, heap=heap),
                                        refs=_type_refs,
                                        both=False, kind=k, type=t)
        elif isbuiltin(t) and t not in _typedefs:  # array, range, xrange in Python 2.x
            _typedefs[t] = _Typedef(base=_basicsize(t, base=base),
                                    both=False, kind=_kind_ignored, type=t)
        else:
            raise KeyError('asizeof typedef %r bad: %r %r' % (self, (c, k), self.both))

    def set(self, safe_len=False, **kwds):
        '''Set one or more attributes.
        '''
        if kwds:  # double check
            d = self.kwds()
            d.update(kwds)
            self.reset(**d)
        if safe_len and self.item:
            self.leng = _len

    def reset(self, base=0, item=0, leng=None, refs=None,
                                    both=True, kind=None, type=None):
        '''Reset all specified attributes.
        '''
        if base < 0:
            raise ValueError('invalid option: %s=%r' % ('base', base))
        else:
            self.base = base
        if item < 0:
            raise ValueError('invalid option: %s=%r' % ('item', item))
        else:
            self.item = item
        if leng in _all_lengs:  # XXX or _callable(leng)
            self.leng = leng
        else:
            raise ValueError('invalid option: %s=%r' % ('leng', leng))
        if refs in _all_refs:  # XXX or _callable(refs)
            self.refs = refs
        else:
            raise ValueError('invalid option: %s=%r' % ('refs', refs))
        if both in (False, True):
            self.both = both
        else:
            raise ValueError('invalid option: %s=%r' % ('both', both))
        if kind in _all_kinds:
            self.kind = kind
        else:
            raise ValueError('invalid option: %s=%r' % ('kind', kind))
        self.type = type

_typedefs = {}  # [key] = _Typedef()

def _typedef_both(t, base=0, item=0, leng=None, refs=None, kind=_kind_static, heap=False):
    '''Add new typedef for both data and code.
    '''
    v = _Typedef(base=_basicsize(t, base=base), item=_itemsize(t, item),
                 refs=refs, leng=leng,
                 both=True, kind=kind, type=t)
    v.save(t, base=base, heap=heap)
    return v  # for _dict_typedef

def _typedef_code(t, base=0, refs=None, kind=_kind_static, heap=False):
    '''Add new typedef for code only.
    '''
    v = _Typedef(base=_basicsize(t, base=base),
                 refs=refs,
                 both=False, kind=kind, type=t)
    v.save(t, base=base, heap=heap)
    return v  # for _dict_typedef

 # static typedefs for data and code types
_typedef_both(complex)
_typedef_both(float)
_typedef_both(list,     refs=_seq_refs, leng=_len_list, item=_sizeof_Cvoidp)  # sizeof(PyObject*)
_typedef_both(tuple,    refs=_seq_refs, leng=_len,      item=_sizeof_Cvoidp)  # sizeof(PyObject*)
_typedef_both(property, refs=_prop_refs)
_typedef_both(type(Ellipsis))
_typedef_both(type(None))

 # _Slots is a special tuple, see _Slots.__doc__
_typedef_both(_Slots, item=_sizeof_Cvoidp,
                      leng=_len_slots,  # length less one
                      refs=None,  # but no referents
                      heap=True)  # plus head

 # dict, dictproxy, dict_proxy and other dict-like types
_dict_typedef = _typedef_both(dict,        item=_sizeof_CPyDictEntry, leng=_len_dict, refs=_dict_refs)
try:  # <type dictproxy> only in Python 2.x
    _typedef_both(Types.DictProxyType,     item=_sizeof_CPyDictEntry, leng=_len_dict, refs=_dict_refs)
except AttributeError:  # XXX any class __dict__ is <type dict_proxy> in Python 3.0?
    _typedef_both(type(_Typedef.__dict__), item=_sizeof_CPyDictEntry, leng=_len_dict, refs=_dict_refs)
 # other dict-like classes and types may be derived or inferred,
 # provided the module and class name is listed here (see functions
 # adict, _isdictclass and _infer_dict for further details)
_dict_classes = {'UserDict': ('IterableUserDict',  'UserDict'),
                 'weakref' : ('WeakKeyDictionary', 'WeakValueDictionary')}
try:  # <type module> is essentially a dict
    _typedef_both(Types.ModuleType, base=_dict_typedef.base,
                                    item=_dict_typedef.item + _sizeof_CPyModuleObject,
                                    leng=_len_module, refs=_module_refs)
except AttributeError:  # missing
    pass

 # newer or obsolete types
try:
    from array import array  # array type
    _typedef_both(array, leng=_len_array, item=_sizeof_Cbyte)
except ImportError:  # missing
    pass

try:  # bool has non-zero __itemsize__ in 3.0
    _typedef_both(bool)
except NameError:  # missing
    pass

try:  # ignore basestring
    _typedef_both(basestring, leng=None)
except NameError:  # missing
    pass

try:
    if isbuiltin(buffer):  # Python 2.2
        _typedef_both(type(buffer('')), item=_sizeof_Cbyte, leng=_len)  # XXX len in bytes?
    else:
        _typedef_both(buffer,           item=_sizeof_Cbyte, leng=_len)  # XXX len in bytes?
except NameError:  # missing
    pass

try:
    _typedef_both(bytearray, item=_sizeof_Cbyte, leng=_len_bytearray)  #PYCHOK bytearray new in 2.6, 3.0
except NameError:  # missing
    pass
try:
    if type(bytes) is not type(str):  # bytes is str in 2.6 #PYCHOK bytes new in 2.6, 3.0
      _typedef_both(bytes, item=_sizeof_Cbyte, leng=_len)  #PYCHOK bytes new in 2.6, 3.0
except NameError:  # missing
    pass
try:  # XXX like bytes
    _typedef_both(str8, item=_sizeof_Cbyte, leng=_len)  #PYCHOK str8 new in 2.6, 3.0  # pyflakes:ignore
except NameError:  # missing
    pass

try:
    _typedef_both(enumerate, refs=_enum_refs)
except NameError:  # missing
    pass

try:  # Exception is type in Python 3.0
    _typedef_both(Exception, refs=_exc_refs)
except:  # missing
    pass  #PYCHOK OK

try:
    _typedef_both(file, refs=_file_refs)
except NameError:  # missing
    pass

try:
    _typedef_both(frozenset, item=_sizeof_Csetentry, leng=_len_set, refs=_seq_refs)
except NameError:  # missing
    pass
try:
    _typedef_both(set,       item=_sizeof_Csetentry, leng=_len_set, refs=_seq_refs)
except NameError:  # missing
    pass

try:  # not callable()
    _typedef_both(Types.GetSetDescriptorType)
except AttributeError:  # missing
    pass

try:  # if long exists, it is multi-precision ...
    _typedef_both(long, item=_sizeof_Cdigit, leng=_len_int)
    _typedef_both(int)  # ... and int is fixed size
except NameError:  # no long, only multi-precision int in Python 3.0
    _typedef_both(int,  item=_sizeof_Cdigit, leng=_len_int)

try:  # not callable()
    _typedef_both(Types.MemberDescriptorType)
except AttributeError:  # missing
    pass

try:
    _typedef_both(type(NotImplemented))  # == Types.NotImplementedType
except NameError:  # missing
    pass

try:
    _typedef_both(range)
except NameError:  # missing
    pass
try:
    _typedef_both(xrange)
except NameError:  # missing
    pass

try:
    _typedef_both(reversed, refs=_enum_refs)
except NameError:  # missing
    pass

try:
    _typedef_both(slice, item=_sizeof_Cvoidp, leng=_len_slice)  # XXX worst-case itemsize?
except NameError:  # missing
    pass

try:
    from os import curdir, stat, statvfs
    _typedef_both(type(stat(   curdir)), refs=_stat_refs)     # stat_result
    _typedef_both(type(statvfs(curdir)), refs=_statvfs_refs,  # statvfs_result
                                         item=_sizeof_Cvoidp, leng=_len)
except ImportError:  # missing
    pass

try:
    from struct import Struct  # only in Python 2.5 and 3.0
    _typedef_both(Struct, item=_sizeof_Cbyte, leng=_len_struct)  # len in bytes
except ImportError:  # missing
    pass

try:
    _typedef_both(Types.TracebackType, refs=_tb_refs)
except AttributeError:  # missing
    pass

try:
    _typedef_both(unicode, leng=_len_unicode, item=_sizeof_Cunicode)
    _typedef_both(str,     leng=_len,         item=_sizeof_Cbyte)  # 1-byte char
except NameError:  # str is unicode
    _typedef_both(str,     leng=_len_unicode, item=_sizeof_Cunicode)

try:  # <type 'KeyedRef'>
    _typedef_both(Weakref.KeyedRef, refs=_weak_refs, heap=True)  # plus head
except AttributeError:  # missing
    pass

try:  # <type 'weakproxy'>
    _typedef_both(Weakref.ProxyType)
except AttributeError:  # missing
    pass

try:  # <type 'weakref'>
    _typedef_both(Weakref.ReferenceType, refs=_weak_refs)
except AttributeError:  # missing
    pass

 # some other, callable types
_typedef_code(object,     kind=_kind_ignored)
_typedef_code(super,      kind=_kind_ignored)
_typedef_code(_Type_type, kind=_kind_ignored)

try:
    _typedef_code(classmethod, refs=_im_refs)
except NameError:
    pass
try:
    _typedef_code(staticmethod, refs=_im_refs)
except NameError:
    pass
try:
    _typedef_code(Types.MethodType, refs=_im_refs)
except NameError:
    pass

try:  # generator, code only, no len(), not callable()
    _typedef_code(Types.GeneratorType, refs=_gen_refs)
except AttributeError:  # missing
    pass

try:  # <type 'weakcallableproxy'>
    _typedef_code(Weakref.CallableProxyType, refs=_weak_refs)
except AttributeError:  # missing
    pass

 # any type-specific iterators
s = [_items({}), _keys({}), _values({})]
try:  # reversed list and tuples iterators
    s.extend([reversed([]), reversed(())])
except NameError:  # missing
    pass
try:  # range iterator
    s.append(xrange(1))
except NameError:  # missing
    pass
try:  # callable-iterator
    from re import finditer
    s.append(finditer('', ''))
except ImportError:  # missing
    pass
for t in _values(_typedefs):
    if t.type and t.leng:
        try:  # create an (empty) instance
            s.append(t.type())
        except TypeError:
            pass
for t in s:
    try:
        i = iter(t)
        _typedef_both(type(i), leng=_len_iter, refs=_iter_refs, item=0)  # no itemsize!
    except (KeyError, TypeError):  # ignore non-iterables, duplicates, etc.
        pass
del i, s, t


def _typedef(obj, derive=False, infer=False):
    '''Create a new typedef for an object.
    '''
    t =  type(obj)
    v = _Typedef(base=_basicsize(t, obj=obj),
                 kind=_kind_dynamic, type=t)
  ##_printf('new %r %r/%r %s', t, _basicsize(t), _itemsize(t), _repr(dir(obj)))
    if ismodule(obj):  # handle module like dict
        v.dup(item=_dict_typedef.item + _sizeof_CPyModuleObject,
              leng=_len_module,
              refs=_module_refs)
    elif isframe(obj):
        v.set(base=_basicsize(t, base=_sizeof_CPyFrameObject, obj=obj),
              item=_itemsize(t),
              leng=_len_frame,
              refs=_frame_refs)
    elif iscode(obj):
        v.set(base=_basicsize(t, base=_sizeof_CPyCodeObject, obj=obj),
              item=_sizeof_Cvoidp,
              leng=_len_code,
              refs=_co_refs,
              both=False)  # code only
    elif _callable(obj):
        if isclass(obj):  # class or type
            v.set(refs=_class_refs,
                  both=False)  # code only
            if obj.__module__ in _builtin_modules:
                v.set(kind=_kind_ignored)
        elif isbuiltin(obj):  # function or method
            v.set(both=False,  # code only
                  kind=_kind_ignored)
        elif isfunction(obj):
            v.set(refs=_func_refs,
                  both=False)  # code only
        elif ismethod(obj):
            v.set(refs=_im_refs,
                  both=False)  # code only
        elif isclass(t):  # callable instance, e.g. SCons,
             # handle like any other instance further below
            v.set(item=_itemsize(t), safe_len=True,
                  refs=_inst_refs)  # not code only!
        else:
            v.set(both=False)  # code only
    elif _issubclass(t, dict):
        v.dup(kind=_kind_derived)
    elif _isdictclass(obj) or (infer and _infer_dict(obj)):
        v.dup(kind=_kind_inferred)
    elif getattr(obj, '__module__', None) in _builtin_modules:
        v.set(kind=_kind_ignored)
    else:  # assume an instance of some class
        if derive:
            p = _derive_typedef(t)
            if p:  # duplicate parent
                v.dup(other=p, kind=_kind_derived)
                return v
        if _issubclass(t, Exception):
            v.set(item=_itemsize(t), safe_len=True,
                  refs=_exc_refs,
                  kind=_kind_derived)
        elif isinstance(obj, Exception):
            v.set(item=_itemsize(t), safe_len=True,
                  refs=_exc_refs)
        else:
            v.set(item=_itemsize(t), safe_len=True,
                  refs=_inst_refs)
    return v


class _Prof(object):
    '''Internal type profile class.
    '''
    total  = 0     # total size
    high   = 0     # largest size
    number = 0     # number of (unique) objects
    objref = None  # largest object (weakref)
    weak   = False # objref is weakref(object)

    def __cmp__(self, other):
        if self.total < other.total:
            return -1
        if self.total > other.total:
            return +1
        if self.number < other.number:
            return -1
        if self.number > other.number:
            return +1
        return 0

    def __lt__(self, other):  # for Python 3.0
        return self.__cmp__(other) < 0

    def format(self, clip=0, grand=None):
        '''Return format dict.
        '''
        if self.number > 1:  # avg., plural
            a, p = int(self.total / self.number), 's'
        else:
            a, p = self.total, ''
        o = self.objref
        if self.weak:  # weakref'd
            o = o()
        t = _SI2(self.total)
        if grand:
            t += ' (%s)' % _p100(self.total, grand, prec=0)
        return _kwds(avg=_SI2(a),         high=_SI2(self.high),
                     lengstr=_lengstr(o), obj=_repr(o, clip=clip),
                     plural=p,            total=t)

    def update(self, obj, size):
        '''Update this profile.
        '''
        self.number += 1
        self.total  += size
        if self.high < size:  # largest
           self.high = size
           try:  # prefer using weak ref
               self.objref, self.weak = Weakref.ref(obj), True
           except TypeError:
               self.objref, self.weak = obj, False


 # public classes

class Asized(object):
    '''Store the results of an  asized object
       in these 4 attributes:

       size - total size of the object
       flat - flat size of the object
       name - name or repr of the object
       refs - tuple containing an instance
              of  Asized for each referent
    '''
    strf = (    '%s',  # default name format
            '[K] %s',  # dict key, see _dict_refs
            '[V] %s')  # dict value, see _dict_refs

    def __init__(self, size, flat, refs=(), name=None):
        self.size = size  # total size
        self.flat = flat  # flat size
        self.name = name  # name, repr or None
        self.refs = tuple(refs)

    def __str__(self):
        return 'size %r, flat %r, refs[%d], name %r' % (
                self.size, self.flat, len(self.refs), self.name)

    def format(self, named):
        '''Format name from _NamedRef instance.
        '''
        return self.strf[named.typ] % named.name

class Asizer(object):
    '''Sizer state and options.
    '''
    _align_    = 8
    _all_      = False
    _clip_     = 80
    _code_     = False
    _derive_   = False
    _detail_   = 0  # for Asized only
    _infer_    = False
    _limit_    = 100
    _stats_    = 0

    _cutoff    = 0  # in percent
    _depth     = 0  # recursion depth
    _duplicate = 0
    _excl_d    = None  # {}
    _ign_d     = _kind_ignored
    _incl      = ''  # or ' (incl. code)'
    _mask      = 7   # see _align_
    _missed    = 0   # due to errors
    _profile   = False
    _profs     = None  # {}
    _seen      = None  # {}
    _total     = 0   # total size

    def __init__(self, **opts):
        '''See method  reset for the available options.
        '''
        self._excl_d = {}
        self.reset(**opts)

    def _clear(self):
        '''Clear state.
        '''
        self._all_      = False
        self._depth     = 0   # recursion depth
        self._duplicate = 0
        self._incl      = ''  # or ' (incl. code)'
        self._missed    = 0   # due to errors
        self._profile   = False
        self._profs     = {}
        self._seen      = {}
        self._total     = 0   # total size
        for k in _keys(self._excl_d):
            self._excl_d[k] = 0

    def _nameof(self, obj):
        '''Return the object's name.
        '''
        return _nameof(obj, '') or self._repr(obj)

    def _prepr(self, obj):
        '''Like prepr().
        '''
        return _prepr(obj, clip=self._clip_)

    def _prof(self, key):
        '''Get _Prof object.
        '''
        p = self._profs.get(key, None)
        if not p:
            self._profs[key] = p = _Prof()
        return p

    def _repr(self, obj):
        '''Like repr().
        '''
        return _repr(obj, clip=self._clip_)

    def _sizer(self, obj, deep, sized):
        '''Size an object, recursively.
        '''
        s, f, i = 0, 0, id(obj)
         # skip obj if seen before
         # or if ref of a given obj
        if i in self._seen:
            if deep:
                self._seen[i] += 1
                if sized:
                    s = sized(s, f, name=self._nameof(obj))
                return s
        else:
            self._seen[i] = 0
        try:
            k, rs = _objkey(obj), []
            if k in self._excl_d:
                self._excl_d[k] += 1
            else:
                v = _typedefs.get(k, None)
                if not v:  # new typedef
                    _typedefs[k] = v = _typedef(obj, derive=self._derive_,
                                                     infer=self._infer_)
                if (v.both or self._code_) and v.kind is not self._ign_d:
                    s = f = v.flat(obj, self._mask)  # flat size
                    if self._profile:  # profile type
                        self._prof(k).update(obj, s)
                     # recurse, but not for nested modules
                    if deep < self._limit_ and not (deep and ismodule(obj)):
                         # add sizes of referents
                        r, z, d = v.refs, self._sizer, deep + 1
                        if self._all_:  # use 'all' referents
                            r = _getreferents(obj)
                            if r:
                                t = id(r)
                                if t in self._seen:
                                    for o in r:  # no sum(<generator_expression>) in Python 2.2
                                        s += z(o, d, None)
                                else:  # exclude container
                                    self._seen[t] = 0
                                    for o in r:  # no sum(<generator_expression>) in Python 2.2
                                        s += z(o, d, None)
                                    del self._seen[t]
                        elif r:  # and _callable(r):
                            if sized and deep < self._detail_:
                                  # use named referents
                                for o in r(obj, True):
                                    if isinstance(o, _NamedRef):
                                        t = z(o.ref, d, sized)
                                        t.name = t.format(o)
                                    else:
                                        t = z(o, d, sized)
                                        t.name = self._nameof(o)
                                    rs.append(t)
                                    s += t.size
                            else:  # no sum(<generator_expression>) in Python 2.2
                                for o in r(obj, False):
                                    s += z(o, d, None)
                         # recursion depth
                        if self._depth < d:
                           self._depth = d
            self._seen[i] += 1
        except RuntimeError:  # XXX RecursionLimitExceeded:
            self._missed += 1
        if sized:
            s = sized(s, f, name=self._nameof(obj), refs=rs)
        return s

    def _sizes(self, objs, sized=None):
        '''Return the size or an Asized instance for each
           given object and the total size.  The total
           includes the size of duplicates only once.
        '''
        self.exclude_refs(*objs)  # skip refs to objs
        s, t = {}, []
        for o in objs:
            i = id(o)
            if i in s:  # duplicate
                self._seen[i] += 1
                self._duplicate += 1
            else:
                s[i] = self._sizer(o, 0, sized)
            t.append(s[i])
        if sized:
            s = _sum([i.size for i in _values(s)])  # [] for Python 2.2
        else:
            s = _sum(_values(s))
        self._total += s  # accumulate
        return s, tuple(t)

    def asized(self, *objs, **opts):
        '''Size each object and return an Asized instance with
           size information and referents up to the given detail
           level (and with modified options, see method  set).

           If only one object is given, the return value is the
           Asized instance for that object.
        '''
        if opts:
            self.set(**opts)
        if self._all_:
            raise KeyError('invalid option: %s=%r' % ('all', self._all_))
        _, t = self._sizes(objs, Asized)
        if len(t) == 1:
            t = t[0]
        return t

    def asizeof(self, *objs, **opts):
        '''Return the combined size of the given objects
           (with modified options, see also method  set).
        '''
        if opts:
            self.set(**opts)
        s, _ = self._sizes(objs, None)
        return s

    def asizesof(self, *objs, **opts):
        '''Return the individual sizes of the given objects
           (with modified options, see also method  set).
        '''
        if opts:
            self.set(**opts)
        _, t = self._sizes(objs, None)
        return t

    def exclude_refs(self, *objs):
        '''Exclude any references to the specified objects from sizing.

           While any references to the given objects are excluded, the
           objects will be sized if specified as positional arguments
           in subsequent calls to methods  asizeof and  asizesof.
        '''
        for o in objs:
            self._seen.setdefault(id(o), 0)

    def exclude_types(self, *objs):
        '''Exclude the specified object instances and types from sizing.

           All instances and types of the given objects are excluded,
           even objects specified as positional arguments in subsequent
           calls to methods  asizeof and  asizesof.
        '''
        for o in objs:
            for t in _keytuple(o):
                if t and t not in self._excl_d:
                    self._excl_d[t] = 0

    def print_profiles(self, w=0, cutoff=0, **print3opts):
        '''Print the profiles above cutoff percentage.

               w=0            -- indentation for each line
               cutoff=0       -- minimum percentage printed
               print3options  -- print options, as in Python 3.0
        '''
         # get the profiles with non-zero size or count
        t = [(v, k) for k, v in _items(self._profs) if v.total > 0 or v.number > 1]
        if (len(self._profs) - len(t)) < 9:  # just show all
            t = [(v, k) for k, v in _items(self._profs)]
        if t:
            s = ''
            if self._total:
                s = ' (% of grand total)'
                c = max(cutoff, self._cutoff)
                c = int(c * 0.01 * self._total)
            else:
                c = 0
            _printf('%s%*d profile%s:  total%s, average, and largest flat size%s:  largest object',
                     linesep, w, len(t), _plural(len(t)), s, self._incl, **print3opts)
            r = len(t)
            for v, k in _sorted(t, reverse=True):
                s = 'object%(plural)s:  %(total)s, %(avg)s, %(high)s:  %(obj)s%(lengstr)s' % v.format(self._clip_, self._total)
                _printf('%*d %s %s', w, v.number, self._prepr(k), s, **print3opts)
                r -= 1
                if r > 1 and v.total < c:
                    c = max(cutoff, self._cutoff)
                    _printf('%+*d profiles below cutoff (%.0f%%)', w, r, c)
                    break
            z = len(self._profs) - len(t)
            if z > 0:
                _printf('%+*d %r object%s', w, z, 'zero', _plural(z), **print3opts)

    def print_stats(self, objs=(), opts={}, sized=(), sizes=(), stats=3.0, **print3opts):
        '''Print the statistics.

               w=0            -- indentation for each line
               objs=()        -- optional, list of objects
               opts={}        -- optional, dict of options used
               sized=()       -- optional, tuple of  Asized instances returned
               sizes=()       -- optional, tuple of sizes returned
               stats=3.0      -- print statistics and cutoff percentage
               print3options  -- print options, as in Python 3.0
        '''
        s = min(opts.get('stats', stats) or 0, self._stats_)
        if s > 0:  # print stats
            t = self._total + self._missed + _sum(_values(self._seen))
            w = len(str(t)) + 1
            t = c = ''
            o = _kwdstr(**opts)
            if o and objs:
                c = ', '
             # print header line(s)
            if sized and objs:
                n = len(objs)
                if n > 1:
                    _printf('%sasized(...%s%s) ...', linesep, c, o, **print3opts)
                    for i in range(n):  # no enumerate in Python 2.2.3
                        _printf('%*d: %s', w-1, i, sized[i], **print3opts)
                else:
                    _printf('%sasized(%s): %s', linesep, o, sized, **print3opts)
            elif sizes and objs:
                _printf('%sasizesof(...%s%s) ...', linesep, c, o, **print3opts)
                for z, o in zip(sizes, objs):
                    _printf('%*d bytes%s%s:  %s', w, z, _SI(z), self._incl, self._repr(o), **print3opts)
            else:
                if objs:
                    t = self._repr(objs)
                _printf('%sasizeof(%s%s%s) ...', linesep, t, c, o, **print3opts)
             # print summary
            self.print_summary(w=w, objs=objs, **print3opts)
            if s > 1:  # print profile
                c = int(s - int(s)) * 100
                self.print_profiles(w=w, cutoff=c, **print3opts)
                if s > 2:  # print typedefs
                    self.print_typedefs(w=w, **print3opts)

    def print_summary(self, w=0, objs=(), **print3opts):
        '''Print the summary statistics.

               w=0            -- indentation for each line
               objs=()        -- optional, list of objects
               print3options  -- print options, as in Python 3.0
        '''
        _printf('%*d bytes%s%s', w, self._total, _SI(self._total), self._incl, **print3opts)
        if self._mask:
            _printf('%*d byte aligned', w, self._mask + 1, **print3opts)
        _printf('%*d byte sizeof(void*)', w, _sizeof_Cvoidp, **print3opts)
        n = len(objs or ())
        if n > 0:
            d = self._duplicate or ''
            if d:
                d = ', %d duplicate' % self._duplicate
            _printf('%*d object%s given%s', w, n, _plural(n), d, **print3opts)
        t = _sum([1 for t in _values(self._seen) if t != 0])  # [] for Python 2.2
        _printf('%*d object%s sized', w, t, _plural(t), **print3opts)
        if self._excl_d:
            t = _sum(_values(self._excl_d))
            _printf('%*d object%s excluded', w, t, _plural(t), **print3opts)
        t = _sum(_values(self._seen))
        _printf('%*d object%s seen', w, t, _plural(t), **print3opts)
        if self._missed > 0:
            _printf('%*d object%s missed', w, self._missed, _plural(self._missed), **print3opts)
        if self._depth > 0:
            _printf('%*d recursion depth', w, self._depth, **print3opts)

    def print_typedefs(self, w=0, **print3opts):
        '''Print the types and dict tables.

               w=0            -- indentation for each line
               print3options  -- print options, as in Python 3.0
        '''
        for k in _all_kinds:
             # XXX Python 3.0 doesn't sort type objects
            t = [(self._prepr(a), v) for a, v in _items(_typedefs) if v.kind == k and (v.both or self._code_)]
            if t:
                _printf('%s%*d %s type%s:  basicsize, itemsize, _len_(), _refs()',
                         linesep, w, len(t), k, _plural(len(t)), **print3opts)
                for a, v in _sorted(t):
                    _printf('%*s %s:  %s', w, '', a, v, **print3opts)
         # dict and dict-like classes
        t = _sum([len(v) for v in _values(_dict_classes)])  # [] for Python 2.2
        if t:
            _printf('%s%*d dict/-like classes:', linesep, w, t, **print3opts)
            for m, v in _items(_dict_classes):
                _printf('%*s %s:  %s', w, '', m, self._prepr(v), **print3opts)

    def set(self, align=None, code=None, detail=None, limit=None, stats=None):
        '''Set some options.  Any options not set
           remain the same as the previous setting.

               align=8     -- size alignment
               code=False  -- incl. (byte)code size
               detail=0    -- Asized refs level
               limit=100   -- recursion limit
               stats=0.0   -- print statistics and cutoff percentage
        '''
         # adjust
        if align is not None:
            self._align_ = align
            if align > 1:
                self._mask = align - 1
                if (self._mask & align) != 0:
                    raise ValueError('invalid option: %s=%r' % ('align', align))
            else:
                self._mask = 0
        if code is not None:
            self._code_ = code
            if code:  # incl. (byte)code
                self._incl = ' (incl. code)'
        if detail is not None:
            self._detail_ = detail
        if limit is not None:
            self._limit_ = limit
        if stats is not None:
            self._stats_ = s = int(stats)
            self._cutoff = (stats - s) * 100
            if s > 1:  # profile types
                self._profile = True
            else:
                self._profile = False

    def _get_duplicate(self):
        '''Number of duplicate objects.
        '''
        return self._duplicate
    duplicate = property(_get_duplicate, doc=_get_duplicate.__doc__)

    def _get_missed(self):
        '''Number of objects missed due to errors.
        '''
        return self._missed
    missed = property(_get_missed, doc=_get_missed.__doc__)

    def _get_total(self):
        '''Total size accumulated so far.
        '''
        return self._total
    total = property(_get_total, doc=_get_total.__doc__)

    def reset(self, align=8,  all=False,    clip=80,     code=False, derive=False,  #PYCHOK expected
                    detail=0, ignored=True, infer=False, limit=100,  stats=0):
        '''Reset options, state, etc.

        The available options and default values are:

             align=8       -- size alignment
             all=False     -- all current GC objects and referents
             clip=80       -- clip repr() strings
             code=False    -- incl. (byte)code size
             derive=False  -- derive from super type
             detail=0      -- Asized refs level
             ignored=True  -- ignore certain types
             infer=False   -- try to infer types
             limit=100     -- recursion limit
             stats=0.0     -- print statistics and cutoff percentage

        See function  asizeof for a description of the options.
        '''
         # options
        self._align_  = align
        self._all_    = all
        self._clip_   = clip
        self._code_   = code
        self._derive_ = derive
        self._detail_ = detail  # for Asized only
        self._infer_  = infer
        self._limit_  = limit
        self._stats_  = stats
        if ignored:
            self._ign_d = _kind_ignored
        else:
            self._ign_d = None
         # clear state
        self._clear()
        self.set(align=align, code=code, stats=stats)


 # public functions

def adict(*classes):
    '''Install one or more classes to be handled as dict.
    '''
    a = True
    for c in classes:
         # if class is dict-like, add class
         # name to _dict_classes[module]
        if isclass(c) and _infer_dict(c):
            t = _dict_classes.get(c.__module__, ())
            if c.__name__ not in t:  # extend tuple
                _dict_classes[c.__module__] = t + (c.__name__,)
        else:  # not a dict-like class
            a = False
    return a  # all installed if True

_asizer = Asizer()

def asized(*objs, **opts):
    '''Return a tuple containing an  Asized instance for each
       object passed as positional argment using the following
       options.

           align=8       -- size alignment
           all=False     -- all current GC objects and referents
           clip=80       -- clip repr() strings
           code=False    -- incl. (byte)code size
           derive=False  -- derive from super type
           detail=0      -- Asized refs level
           ignored=True  -- ignore certain types
           infer=False   -- try to infer types
           limit=100     -- recursion limit
           stats=0.0     -- print statistics and cutoff percentage

       If only one object is given, the return value is the Asized
       instance for that object.

       Set  detail to the desired referents level (recursion depth).

       See function  asizeof for descriptions of the other options.

       The length of the returned tuple matches the number of given
       objects, if more than one object is given.
    '''
    t = _objs(objs, **opts)
    if t:
        _asizer.reset(**opts)
        s = _asizer.asized(*t)
        _asizer.print_stats(objs=t, opts=opts, sized=s)
        _asizer._clear()
    else:
        s = ()
    return s

def asizeof(*objs, **opts):
    '''Return the combined size in bytes of all objects passed
       as positional argments.

       The available options and defaults are the following.

           align=8       -- size alignment
           all=False     -- all current GC objects and referents
           clip=80       -- clip ``repr()`` strings
           code=False    -- incl. (byte)code size
           derive=False  -- derive from super type
           ignored=True  -- ignore certain types
           infer=False   -- try to infer types
           limit=100     -- recursion limit
           stats=0.0     -- print statistics and cutoff percentage

       Set  align to a power of 2 to align sizes.  Any value less
       than 2 avoids size alignment.

       All current GC objects are sized if  all is True and if no
       positional arguments are supplied.  Also, if  all is True
       the GC referents are used instead of the limited ones.

       A positive  clip value truncates all repr() strings to at
       most  clip characters.

       The (byte)code size of callable objects like functions,
       methods, classes, etc. is included only if  code is True.

       If  derive is True, new types are handled like an existing
       (super) type provided there is one and only of those.

       By default, certain base types like object are ignored for
       sizing.  Set  ignored to False to force all ignored types
       in the size of objects.

       By default certain base types like object, super, etc. are
       ignored.  Set  ignored to False to include those.

       If  infer is True, new types are inferred from attributes
       (only implemented for dict types on callable attributes
       as get, has_key, items, keys and values).

       Set  limit to a positive value to accumulate the sizes of
       the referents of each object, recursively up to the limit.
       Using  limit zero returns the sum of the flat [1] sizes of
       the given objects.  High  limit values may cause runtime
       errors and miss objects for sizing.

       A positive value for  stats prints up to 8 statistics, (1)
       a summary of the number of objects sized and seen, (2) a
       simple profile of the sized objects by type and (3+) up to
       6 tables showing the static, dynamic, derived, ignored,
       inferred and dict types used, found respectively installed.

       The fractional part of the  stats value (x 100) is the cutoff
       percentage for simple profiles.  Objects below the cutoff
       value are not reported.

       [1] See the documentation of this module for the definition
           of flat size.
    '''
    t = _objs(objs, **opts)
    if t:
        _asizer.reset(**opts)
        s = _asizer.asizeof(*t)
        _asizer.print_stats(objs=t, opts=opts)
        _asizer._clear()
    else:
        s = 0
    return s

def asizesof(*objs, **opts):
    '''Return a tuple containing the size in bytes of all objects
       passed as positional argments using the following options.

           align=8       -- size alignment
           all=False     -- use GC objects and referents
           clip=80       -- clip ``repr()`` strings
           code=False    -- incl. (byte)code size
           derive=False  -- derive from super type
           ignored=True  -- ignore certain types
           infer=False   -- try to infer types
           limit=100     -- recursion limit
           stats=0.0     -- print statistics and cutoff percentage

       See function  asizeof for a description of the options.

       The length of the returned tuple equals the number of given
       objects.
    '''
    t = _objs(objs, **opts)
    if t:
        _asizer.reset(**opts)
        s = _asizer.asizesof(*t)
        _asizer.print_stats(objs=t, opts=opts, sizes=s)
        _asizer._clear()
    else:
        s = ()
    return s

def _typedefof(obj, save=False, **opts):
    '''Get the typedef for an object.
    '''
    k = _objkey(obj)
    v = _typedefs.get(k, None)
    if not v:  # new typedef
        v = _typedef(obj, **opts)
        if save:
            _typedefs[k] = v
    return v

def basicsize(obj, **opts):
    '''Return the basic size of an object (in bytes).

       Valid options and defaults are
           derive=False  -- derive type from super type
           infer=False   -- try to infer types
           save=False    -- save typedef if new
    '''
    v = _typedefof(obj, **opts)
    if v:
        v = v.base
    return v

def flatsize(obj, align=0, **opts):
    '''Return the flat size of an object (in bytes),
       optionally aligned to a given power of 2.

       See function  basicsize for a description of
       the other options.  See the documentation of
       this module for the definition of flat size.
    '''
    v = _typedefof(obj, **opts)
    if v:
        if align > 1:
            m = align - 1
            if (align & m) != 0:
                raise ValueError('invalid option: %s=%r' % ('align', align))
        else:
            m = 0
        v = v.flat(obj, m)
    return v

def itemsize(obj, **opts):
    '''Return the item size of an object (in bytes).

       See function  basicsize for a description of
       the options.
    '''
    v = _typedefof(obj, **opts)
    if v:
        v = v.item
    return v

def leng(obj, **opts):
    '''Return the length of an object (in items).

       See function  basicsize for a description
       of the options.
    '''
    v = _typedefof(obj, **opts)
    if v:
        v = v.leng
        if v and _callable(v):
            v = v(obj)
    return v

def refs(obj, all=False, **opts):
    '''Return (a generator for) specific referents of an
       object.

       If all is True return the GC referents.

       See function  basicsize for a description of the
       options.
    '''
    v = _typedefof(obj, **opts)
    if v:
        if all:  # == True
            v = _getreferents(obj)
        else:
            v = v.refs
            if v and _callable(v):
                v = v(obj, False)
    return v


if __name__ == '__main__':

    argv, MAX = sys.argv, sys.getrecursionlimit()

    def _print_asizeof(obj, infer=False, stats=0):
        a = [_repr(obj),]
        for d, c in ((0, False), (MAX, False), (MAX, True)):
            a.append(asizeof(obj, limit=d, code=c, infer=infer, stats=stats))
        _printf(" asizeof(%s) is %d, %d, %d", *a)

    def _print_functions(obj, name=None, align=8, detail=MAX, code=False, limit=MAX,
                              opt='', **unused):
        if name:
            _printf('%sasizeof functions for %s ... %s', linesep, name, opt)
        _printf('%s(): %s', ' basicsize', basicsize(obj))
        _printf('%s(): %s', ' itemsize',  itemsize(obj))
        _printf('%s(): %r', ' leng',      leng(obj))
        _printf('%s(): %s', ' refs',     _repr(refs(obj)))
        _printf('%s(): %s', ' flatsize',  flatsize(obj, align=align))  # , code=code
        _printf('%s(): %s', ' asized',           asized(obj, align=align, detail=detail, code=code, limit=limit))
      ##_printf('%s(): %s', '.asized',   _asizer.asized(obj, align=align, detail=detail, code=code, limit=limit))

    def _bool(arg):
        a = arg.lower()
        if a in ('1', 't', 'y', 'true', 'yes', 'on'):
            return True
        elif a in ('0', 'f', 'n', 'false', 'no', 'off'):
            return False
        else:
            raise ValueError('bool option expected: %r' % arg)

    def _opts(*opts):
        '''Return True if any oof the given options
           was present in the command line arguments.
        '''
        for o in opts + ('-', '--'):
            if o in argv:
                return True
        return False

    if '-im' in argv or '-import' in argv:
         # import and size modules given as args

        def _aopts(argv, **opts):
            '''Get argv options as typed values.
            '''
            i = 1
            while argv[i].startswith('-'):
                k = argv[i].lstrip('-')
                if 'import'.startswith(k):
                    i += 1
                elif k in opts:
                    t = type(opts[k])
                    if t is bool:
                        t = _bool
                    i += 1
                    opts[k] = t(argv[i])
                    i += 1
                else:
                    raise NameError('invalid option: %s' % argv[i])
            return opts, i

        opts, i = _aopts(argv, align=8, clip=80, code=False, derive=False, detail=MAX, limit=MAX, stats=0)
        while i < len(argv):
            m, i = argv[i], i + 1
            if m == 'eval' and i < len(argv):
                o, i = eval(argv[i]), i + 1
            else:
                o = __import__(m)
            s = asizeof(o, **opts)
            _printf("%sasizeof(%s) is %d", linesep, _repr(o, opts['clip']), s)
            _print_functions(o, **opts)
        argv = []

    elif len(argv) < 2 or _opts('-h', '-help'):
        d = {'-all':               'all=True example',
             '-basic':             'basic examples',
             '-C':                 'Csizeof values',
             '-class':             'class and instance examples',
             '-code':              'code examples',
             '-dict':              'dict and UserDict examples',
           ##'-gc':                'gc examples',
             '-gen[erator]':       'generator examples',
             '-glob[als]':         'globals examples, incl. asized()',
             '-h[elp]':            'print this information',
             '-im[port] <module>': 'imported module example',
             '-int | -long':       'int and long examples',
             '-iter[ator]':        'iterator examples',
             '-loc[als]':          'locals examples',
             '-pair[s]':           'key pair examples',
             '-slots':             'slots examples',
             '-stack':             'stack examples',
             '-sys':               'sys.modules examples',
             '-test':              'test flatsize() vs sys.getsizeof()',
             '-type[def]s':        'type definitions',
             '- | --':             'all examples'}
        w = -max([len(o) for o in _keys(d)])  # [] for Python 2.2
        t = _sorted(['%*s -- %s' % (w, o, t) for o, t in _items(d)])  # [] for Python 2.2
        t = '\n     '.join([''] + t)
        _printf('usage: %s <option> ...\n%s\n', argv[0], t)

    class C: pass

    class D(dict):
        _attr1 = None
        _attr2 = None

    class E(D):
        def __init__(self, a1=1, a2=2):  #PYCHOK OK
            self._attr1 = a1  #PYCHOK OK
            self._attr2 = a2  #PYCHOK OK

    class P(object):
        _p = None
        def _get_p(self):
            return self._p
        p = property(_get_p)  #PYCHOK OK

    class O:  # old style
        a = None
        b = None

    class S(object):  # new style
        __slots__ = ('a', 'b')

    class T(object):
        __slots__ = ('a', 'b')
        def __init__(self):
            self.a = self.b = 0

    if _opts('-all'):  # all=True example
        _printf('%sasizeof(limit=%s, code=%s, %s) ... %s', linesep, 'MAX', True, 'all=True', '-all')
        asizeof(limit=MAX, code=True, stats=MAX, all=True)

    if _opts('-basic'):  # basic examples
        _printf('%sasizeof(%s) for (limit, code) in %s ... %s', linesep, '<basic_objects>', '((0, False), (MAX, False), (MAX, True))', '-basic')
        for o in (None, True, False,
                  1.0, 1.0e100, 1024, 1000000000,
                  '', 'a', 'abcdefg',
                  {}, (), []):
            _print_asizeof(o, infer=True)

    if _opts('-C'):  # show all Csizeof values
        _sizeof_Cdouble  = calcsize('d')  #PYCHOK OK
        _sizeof_Csize_t  = calcsize('Z')  #PYCHOK OK
        _sizeof_Cssize_t = calcsize('z')  #PYCHOK OK
        t = [t for t in locals().items() if t[0].startswith('_sizeof_')]
        _printf('%s%d C sizes: (bytes) ... -C', linesep, len(t))
        for n, v in _sorted(t):
            _printf(' sizeof(%s): %r', n[len('_sizeof_'):], v)

    if _opts('-class'):  # class and instance examples
        _printf('%sasizeof(%s) for (limit, code) in %s ... %s', linesep, '<non-callable>', '((0, False), (MAX, False), (MAX, True))', '-class')
        for o in (C(), C.__dict__,
                  D(), D.__dict__,
                  E(), E.__dict__,
                  P(), P.__dict__, P.p,
                  O(), O.__dict__,
                  S(), S.__dict__,
                  S(), S.__dict__,
                  T(), T.__dict__):
            _print_asizeof(o, infer=True)

    if _opts('-code'):  # code examples
        _printf('%sasizeof(%s) for (limit, code) in %s ... %s', linesep, '<callable>', '((0, False), (MAX, False), (MAX, True))', '-code')
        for o in (C, D, E, P, S, T,  # classes are callable
                  type,
                 _co_refs, _dict_refs, _inst_refs, _len_int, _seq_refs, lambda x: x,
                (_co_refs, _dict_refs, _inst_refs, _len_int, _seq_refs),
                 _typedefs):
            _print_asizeof(o)

    if _opts('-dict'):  # dict and UserDict examples
        _printf('%sasizeof(%s) for (limit, code) in %s ... %s', linesep, '<Dicts>', '((0, False), (MAX, False), (MAX, True))', '-dict')
        try:
            import UserDict  # no UserDict in 3.0
            for o in (UserDict.IterableUserDict(), UserDict.UserDict()):
                _print_asizeof(o)
        except ImportError:
            pass

        class _Dict(dict):
            pass

        for o in (dict(), _Dict(),
                  P.__dict__,  # dictproxy
                  Weakref.WeakKeyDictionary(), Weakref.WeakValueDictionary(),
                 _typedefs):
            _print_asizeof(o, infer=True)

  ##if _opts('-gc'):  # gc examples
      ##_printf('%sasizeof(limit=%s, code=%s, *%s) ...', linesep, 'MAX', False, 'gc.garbage')
      ##from gc import collect, garbage  # list()
      ##asizeof(limit=MAX, code=False, stats=1, *garbage)
      ##collect()
      ##asizeof(limit=MAX, code=False, stats=2, *garbage)

    if _opts('-gen', '-generator'):  # generator examples
        _printf('%sasizeof(%s, code=%s) ... %s', linesep, '<generator>', True, '-gen[erator]')
        def gen(x):
            i = 0
            while i < x:
                yield i
                i += 1
        a = gen(5)
        b = gen(50)
        asizeof(a, code=True, stats=1)
        asizeof(b, code=True, stats=1)
        asizeof(a, code=True, stats=1)

    if _opts('-glob', '-globals'):  # globals examples
        _printf('%sasizeof(%s, limit=%s, code=%s) ... %s', linesep, 'globals()', 'MAX', False, '-glob[als]')
        asizeof(globals(), limit=MAX, code=False, stats=1)
        _print_functions(globals(), 'globals()', opt='-glob[als]')

        _printf('%sasizesof(%s, limit=%s, code=%s) ... %s', linesep, 'globals(), locals()', 'MAX', False, '-glob[als]')
        asizesof(globals(), locals(), limit=MAX, code=False, stats=1)
        asized(globals(), align=0, detail=MAX, limit=MAX, code=False, stats=1)

    if _opts('-int', '-long'):  # int and long examples
        try:
            _L5d  = long(1) << 64
            _L17d = long(1) << 256
            t = '<int>/<long>'
        except NameError:
            _L5d  = 1 << 64
            _L17d = 1 << 256
            t = '<int>'

        _printf('%sasizeof(%s, align=%s, limit=%s) ... %s', linesep, t, 0, 0, '-int')
        for o in (1024, 1000000000,
                  1.0, 1.0e100, 1024, 1000000000,
                  MAX, 1 << 32, _L5d, -_L5d, _L17d, -_L17d):
            _printf(" asizeof(%s) is %s (%s + %s * %s)", _repr(o), asizeof(o, align=0, limit=0),
                                                         basicsize(o), leng(o), itemsize(o))

    if _opts('-iter', '-iterator'):  # iterator examples
        _printf('%sasizeof(%s, code=%s) ... %s', linesep, '<iterator>', False, '-iter[ator]')
        o = iter('0123456789')
        e = iter('')
        d = iter({})
        i = iter(_items({1:1}))
        k = iter(_keys({2:2, 3:3}))
        v = iter(_values({4:4, 5:5, 6:6}))
        l = iter([])
        t = iter(())
        asizesof(o, e, d, i, k, v, l, t, limit=0, code=False, stats=1)
        asizesof(o, e, d, i, k, v, l, t, limit=9, code=False, stats=1)

    if _opts('-loc', '-locals'):  # locals examples
        _printf('%sasizeof(%s, limit=%s, code=%s) ... %s', linesep, 'locals()', 'MAX', False, '-loc[als]')
        asizeof(locals(), limit=MAX, code=False, stats=1)
        _print_functions(locals(), 'locals()', opt='-loc[als]')

    if _opts('-pair', '-pairs'):  # key pair examples
         # <http://jjinux.blogspot.com/2008/08/python-memory-conservation-tip.html>
        _printf('%sasizeof(%s) vs asizeof(%s) ... %s', linesep, 'dict[i][j]', 'dict[(i,j)]', '-pair[s]')
        n = m = 200

        p = {}  # [i][j]
        for i in range(n):
            q = {}
            for j in range(m):
                q[j] = None
            p[i] = q
        p = asizeof(p, stats=1)

        t = {}  # [(i,j)]
        for i in range(n):
            for j in range(m):
                t[(i,j)] = None
        t = asizeof(t, stats=1)

        _printf('%sasizeof(dict[i][j]) is %s of asizeof(dict[(i,j)])', linesep, _p100(p, t))

    if _opts('-slots'):  # slots examples
        _printf('%sasizeof(%s, code=%s) ... %s', linesep, '<__slots__>', False, '-slots')
        class Old:
            pass  # m = None
        class New(object):
            __slots__ = ('n',)
        class Sub(New):  #PYCHOK OK
            __slots__ = {'s': ''}  # duplicate!
            def __init__(self):  #PYCHOK OK
                New.__init__(self)
         # basic instance sizes
        o, n, s = Old(), New(), Sub()
        asizesof(o, n, s, limit=MAX, code=False, stats=1)
         # with unique min attr size
        o.o = 'o'
        n.n = 'n'
        s.n = 'S'
        s.s = 's'
        asizesof(o, n, s, limit=MAX, code=False, stats=1)
         # with duplicate, intern'ed, 1-char string attrs
        o.o = 'x'
        n.n = 'x'
        s.n = 'x'
        s.s = 'x'
        asizesof(o, n, s, 'x', limit=MAX, code=False, stats=1)
         # with larger attr size
        o.o = 'o'*1000
        n.n = 'n'*1000
        s.n = 'n'*1000
        s.s = 's'*1000
        asizesof(o, n, s, 'x'*1000, limit=MAX, code=False, stats=1)

    if _opts('-stack'):  # stack examples
        _printf('%sasizeof(%s, limit=%s, code=%s) ... %s', linesep, 'stack(MAX)', 'MAX', False, '')
        asizeof(stack(MAX), limit=MAX, code=False, stats=1)
        _print_functions(stack(MAX), 'stack(MAX)', opt='-stack')

    if _opts('-sys'):  # sys.modules examples
        _printf('%sasizeof(limit=%s, code=%s, *%s) ... %s', linesep, 'MAX', False, 'sys.modules.values()', '-sys')
        asizeof(limit=MAX, code=False, stats=1, *sys.modules.values())
        _print_functions(sys.modules, 'sys.modules', opt='-sys')

    if _opts('-type', '-types', '-typedefs'):  # show all basic _typedefs
        t = len(_typedefs)
        w = len(str(t)) * ' '
        _printf('%s%d type definitions: basic- and itemsize (leng), kind ... %s', linesep, t, '-type[def]s')
        for k, v in _sorted([(_prepr(k), v) for k, v in _items(_typedefs)]):  # [] for Python 2.2
            s = '%(base)s and %(item)s%(leng)s, %(kind)s%(code)s' % v.format()
            _printf('%s %s: %s', w, k, s)

    if _opts('-test'):
         # compare the results of flatsize() *without* using sys.getsizeof()
         # with the accurate sizes returned by sys.getsizeof() but expect
         # differences for sequences as dicts, lists, sets, tuples, etc.
         # while this is no proof for the accuracy of flatsize() on Python
         # builds without sys.getsizeof(), it does provide some evidence
         # that that flatsize() produces reasonable and usable results
        _printf('%sflatsize() vs sys.getsizeof() ... %s', linesep, '-test')
        t, g, e = [], _getsizeof, 0
        if g:
            for v in _values(_typedefs):
                t.append(v.type)
                try:  # creating one instance
                    if v.type.__module__ not in ('io',):  # avoid 3.0 RuntimeWarning
                        t.append(v.type())
                except (AttributeError, SystemError, TypeError, ValueError):  # ignore errors
                    pass
            t.extend(({1:1, 2:2, 3:3, 4:4, 5:5, 6:6, 7:7, 8:8},
                      [1,2,3,4,5,6,7,8], ['1', '2', '3'], [0] * 100,
                      '12345678', 'x' * 1001,
                      (1,2,3,4,5,6,7,8), ('1', '2', '3'), (0,) * 100,
                      _Slots((1,2,3,4,5,6,7,8)), _Slots(('1', '2', '3')), _Slots((0,) * 100),
                      0, 1 << 8, 1 << 16, 1 << 32, 1 << 64, 1 << 128,
                      complex(0, 1), True, False))
            _getsizeof = None  # zap _getsizeof for flatsize()
            for o in t:
                a = flatsize(o)
                s = sys.getsizeof(o, 0)  # 0 as default #PYCHOK expected
                if a != s:
                     # flatsize approximates the length of sequences
                     # (sys.getsizeof(bool) on 3.0b3 is not correct)
                    if type(o) in (dict, list, set, frozenset, tuple) or (
                       type(o) in (bool,) and sys.version_info[0] == 3):
                        x = 'expected failure'
                    else:
                        x = '%r' % _typedefof(o)
                        e += 1
                    _printf('flatsize() %s vs sys.getsizeof() %s for %s: %s, %s',
                             a, s, _nameof(type(o)), _repr(o), x)
            _getsizeof = g  # restore
        n, p = len(t), 'python %s' % sys.version.split()[0]
        if e:
            _printf('%s%d of %d tests failed or %s on %s', linesep, e, n, _p100(e, n), p)
        elif g:
            _printf('no unexpected failures in %d tests on %s', n, p)
        else:
            _printf('no sys.%s() in this %s', 'getsizeof', p)


# License file from an earlier version of this source file follows:

#---------------------------------------------------------------------
#       Copyright (c) 2002-2009 -- ProphICy Semiconductor, Inc.
#                        All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# - Neither the name of ProphICy Semiconductor, Inc. nor the names
#   of its contributors may be used to endorse or promote products
#   derived from this software without specific prior written
#   permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
#---------------------------------------------------------------------




"""
This class finds the brightest pixel of the image, returns it in white and
other in black.
This class also gives instructions to move the camera.

$Id$
"""



import param
from topo.base.sheet import Sheet
from topo.misc.ptz import PTZ
from topo.base.arrayutil import array_argmax

import numpy as np
from PIL import  Image

try:
    import opencv

except ImportError:
    param.Parameterized().warning("ptztracker.py classes will not be usable; python-opencv is not available.")


class PtzTracker(Sheet):
    """
    Given an incoming Activity pattern, find the brightest pixel and
    output an activity pattern where all but this pixel is set to zero.
    Also controls a pan/tilt/zoom camera, instructing it to move so that
    the brightest pixel will be in the center of the sheet.
    """

    dest_ports=['Activity']
    src_ports=['Activity']

    ratio = param.Number(default=1.33,doc=
        "Ratio depends of the resolution of the camera, here 640x480.")

    fov_x = param.Number(default=76,doc=
        "Field of view along x in degrees, depends of the camera.")

    fov_y = param.Number(default=57,doc=
        """
        Field of view along y in degrees. It is calculate with the resolution
        of the camera, and its field of view. It also depends of the bounds used,
        here the size_normalization used is "fit_shortest". So Sheet coordinates
        along y for camera's image are between -0.5 and 0.5.     
        The original calculation is fov_y=fov_x*0.5*2/ratio.
        """)

    ptz = param.ClassSelector(PTZ,default=None,doc="""
        An instance of ptzcamera.PTZ to be controlled.""")

    # Max Ranges (determined with uvcdynctrl -v -c)
    maxrange_x = param.Number(default=4480, doc=
        """Maximum position of the camera along x.""")

    maxrange_y = param.Number(default=1920, doc=
        """Maximum position of the camera along y.""")   

    #Coordinates in degrees
    y_deg = 0
    x_deg = 0
    #Coordinates in degrees for uvcdynctrl(unit=1/64th of a degree)
    y_deg_uvc = 0
    x_deg_uvc = 0
    #Coordinates of the brightest pixel
    coor = (0,0)
    #Current postitions along x and y
    curr_y=0
    curr_x=0
 
    def __init__(self, **params):
        # Set here to avoid having it one instantiated by default
        if self.ptz==None: self.ptz=PTZ()

    def input_event(self,conn,data):
        self.input_data=data

    def determine_next_position(self,img):
        """
        Determine the next location to move in the specified image, using whatever
        criterion is appropriate for this class.  Returns a tuple of (pos,bbox), 
        where pos is the (row,column) coordinate of the next position, and bbox
        is a bounding box around that coordinate, with whatever size is appropriate
        for this class.  Returns None if no appropriate location can be found.
        """    
        raise NotImplementedError


    def draw_boxes(self,input_data,bboxmin,bboxmax):
        """
        Draws three boxes arond the returned location.
        """
        opencv.cvRectangle(self.input_data, opencv.cvPoint(np.int(bboxmin[0]),np.int(bboxmin[1])),opencv.cvPoint(np.int(bboxmax[0])
            ,np.int(bboxmax[1])),0,1)
        opencv.cvRectangle(self.input_data, opencv.cvPoint(np.int(bboxmin[0])-1,np.int(bboxmin[1])-1),opencv.cvPoint(np.int(bboxmax[0])
            +1,np.int(bboxmax[1])+1),0,1)
        opencv.cvRectangle(self.input_data, opencv.cvPoint(np.int(bboxmin[0])-2,np.int(bboxmin[1])-2),opencv.cvPoint(np.int(bboxmax[0])
            +2,np.int(bboxmax[1])+2),1,1)


    def move_camera(self,pos, bboxmin, bbboxmax,brightpixel):
        """
        Move the camera to centre the returned location.
        """     
        #Determinate the centre of the returned location
        if (self.brightpixel==True):
            self.pos=pos
        else:
            self.pos=(((self.bboxmax[0]+self.bboxmin[0])/2,(self.bboxmax[1]+self.bboxmin[1])/2))
        
        #Find sheet coordinates of the specified position
        self.coor=self.matrixidx2sheet(*self.pos)

        self.y_deg=self.coor[1]*(self.fov_y/2)/0.5
        self.x_deg=self.coor[0]*(self.fov_x/2)/(self.ratio/2)
            
        # The unit of uvcdynctrl is 1/64th of a degree
        ## HACK problem with the scaling. so divide by 2.
        ##it doesn't come from the fov or resolution maybe from somewhere with the coordinates
        self.y_deg_uvc=self.y_deg*64/2
        self.x_deg_uvc=self.x_deg*64/2
           
        self.message("Coordinates of the new position: (%f,%f)" % (self.coor[0],self.coor[1]))
        self.verbose("Current position of the camera (%f,%f) along the two directions of the camera" % (self.curr_x,self.curr_y))
        self.verbose("Movements in degrees of uvcdynctrl (%f,%f)" % (self.x_deg_uvc,self.y_deg_uvc))

         #Use class PTZ in order to move the camera with uvcdynctrl
         # Max Ranges (determined with uvcdynctrl -v -c)
        if ((self.curr_y+self.y_deg_uvc<self.maxrange_y) and (self.curr_y+self.y_deg_uvc>-self.maxrange_y)):
            self.curr_y+=self.y_deg_uvc
            self.ptz.tilt(self.y_deg_uvc)
        else:
            self.message("The camera can't move further along y")
                
        if ((self.curr_x+self.x_deg_uvc<self.maxrange_x) and (self.curr_x+self.x_deg_uvc>-self.maxrange_x)):
            self.curr_x+=self.x_deg_uvc
            self.ptz.pan(self.x_deg_uvc)
        else:
            self.message("The camera can't move further along x")


    def process_current_time(self):
        if hasattr(self, 'input_data'):

            #Find the brightest pixel of the image, put it in white and other pixels in black
            result=self.determine_next_position(self.input_data)
            self.activity*=0

            if result: # Draw a box around the returned location
                self.pos,self.bboxmin,self.bboxmax,self.brightpixel=result
                self.draw_boxes(self.input_data,self.bboxmin,self.bboxmax)

            self.activity+=self.input_data
            self.send_output(src_port='Activity',data=self.activity)
            del self.input_data

            if not result:
                return # Else move camera

            self.move_camera(self.pos,self.bboxmin,self.bboxmax,self.brightpixel)



class BrightPixelTracker(PtzTracker):
    """ 
    This class is used to define the position of the brightest pixel
    in the image.
    """

    def determine_next_position(self,img):
        self.maximum=array_argmax(img)
   
        #Coordinates for the box around the brightest pixel
        self.coormin_bbox=(self.maximum[1]-1,self.maximum[0]-1)
        self.coormax_bbox=(self.maximum[1]+1,self.maximum[0]+1)
        self.brightpixel=True
        return(self.maximum,self.coormin_bbox,self.coormax_bbox,self.brightpixel)



class FaceTracker(PtzTracker):
    """
    This class is used to detect face in the image and draws a
    rectangle around each detected face.
    """

    def determine_next_position(self,image):

        self.image=image
        self.im=array2image(image)
        self.ipl_im = opencv.adaptors.PIL2Ipl(self.im)
        self.storage = opencv.cvCreateMemStorage(0)
        opencv.cvClearMemStorage(self.storage)
        self.cascade = opencv('/usr/share/opencv/haarcascades/haarcascade_frontalface_default.xml',opencv.cvSize(1,1))
        self.faces = opencv.cvHaarDetectObjects(self.ipl_im, self.cascade, self.storage, 1.2, 2,opencv.CV_HAAR_DO_CANNY_PRUNING, opencv.cvSize(50,50))

        if self.faces.total < 1:
            return None
 
        for f in self.faces:
            print "face detected: %s" %f
            #row and column are inverted in Opencv
            self.pos=(f.y,f.x)
            self.coormin_bbox=(self.pos[0],self.pos[1])
            self.coormax_bbox=(self.pos[0]+f.width,self.pos[1]+f.height)
            self.brightpixel=False      
            return (self.pos,self.coormin_bbox,self.coormax_bbox,self.brightpixel)



#Convert an array into PIL image
def array2image(arr):
    arr = arr*255.0
    arr=np.floor(arr)#arr=arr.round()
    arr=arr.astype(np.uint8)
    return Image.fromarray(arr)


__all__ = [
    "PtzTracker",
    "BrightPixelTracker",
    "FaceTracker",
]



"""
Test for the PlotGroup class
$Id$
"""

import unittest
import topo.base.simulation
from pprint import pprint
from topo.plotting import plot
from topo.base.sheet import *
from topo.plotting.plotgroup import *
import numpy.oldnumeric as Numeric
from numpy.oldnumeric import divide

SHOW_PLOTS = False

plotgroup_templates = {}

### JCALERT! This file has to be re-written when the fundamental changes in plot.py
### plotengine.py and plotgroup.py will be finished.
### for the moment, the tests are commented out...
### Also it has to performed the test that were previously performed in testplotengine.py


class TestPlotGroup(unittest.TestCase):

    def setUp(self):
        pass
    
#         pgt = PlotGroupTemplate([('Activity',
#                                   PlotTemplate({'Strength'   : 'Activity',
#                                                 'Hue'        : None,
#                                                 'Confidence' : None}))],
#                                 name='Activity')
#         plotgroup_templates[pgt.name] = pgt
#         pgt = PlotGroupTemplate([('Unit Weights',
#                                   PlotTemplate({'Location'   : (0.0,0.0),
#                                                 'Sheet_name' : 'V1'}))],
#                                 name='Unit Weights')
#         plotgroup_templates[pgt.name] = pgt
#         pgt = PlotGroupTemplate([('Projection',
#                                   PlotTemplate({'Density'         : 25,
#                                                 'Projection_name' : 'None'}))],
#                                 name='Projection')
#         plotgroup_templates[pgt.name] = pgt
#         pgt = PlotGroupTemplate([('Preference',
#                                   PlotTemplate({'Strength'   : 'Activity',
#                                                 'Hue'        : 'Activity',
#                                                 'Confidence' : 'Activity'}))],
#                                 name='Preference Map')
#         plotgroup_templates[pgt.name] = pgt


#     def test_plotgroup_release(self):
#         self.s = Sheet()
#         self.s.activity = Numeric.array([[1,2],[3,4]])
#         # Call s.sheet_view(..) with a parameter
#         sv2 = self.s.sheet_view('Activity')
#         self.s.sheet_views['key']=sv2
#         self.assertEqual(len(self.s.sheet_views.keys()),1)
#         y = plot.Plot(('key',None,None),plot.HSV,self.s)
#         z = plot.Plot(('key',None,None),plot.HSV,self.s)
#         self.pg1 = PlotGroup(plot_list=[y,z])
#         tuples = self.pg1.plots()
#         self.pg1.release_sheetviews()
#         self.assertEqual(len(self.s.sheet_views.keys()),0)
        


#     def test_plotgroup(self):
#         self.s2 = Sheet()
#         ig = ImageGenerator(filename='topo/tests/testsheetview.ppm',
#                          density=100,
#                          bounds=BoundingBox(points=((-0.8,-0.8),(0.8,0.8))))
#         x = plot.Plot((None,None,ig.sheet_view('Activity')),plot.HSV)
#         y = plot.Plot(('Activity',None,None),plot.COLORMAP,ig)
#         z = plot.Plot(('Activity',None,'Activity'),plot.HSV,ig)
#         self.pg1 = PlotGroup(plot_list=[x])
#         self.pg1.add(y)
#         self.pg1.add(z)
#         plot_list = self.pg1.plots()
#         for each in plot_list:
#             (r,g,b) = each.matrices
#             map = RGBMap(r,g,b)
#             if SHOW_PLOTS: map.show()

#     def test_make_plot_group(self):
#         sim = topo.base.simulation.Simulation()
#         pe = topo.plotting.plotengine.PlotEngine(sim)
#       filter = lambda s: True
#         pg = pe.make_plot_group('Activity',
#                                 plotgroup_templates['Activity'],
#                               filter,'BasicPlotGroup')

#     def test_get_plot_group(self):
#         sim = topo.base.simulation.Simulation()
#         pe = topo.plotting.plotengine.PlotEngine(sim)
#         pg = pe.get_plot_group('Activity',
#                                plotgroup_templates['Activity'])
#         pg = pe.get_plot_group('Activity',
#                                plotgroup_templates['Activity'])



#     def test_keyedlist(self):
#         kl = KeyedList()
#         kl = KeyedList(((2,3),(4,5)))
#         self.assertEqual(kl[2],3)
#         self.assertEqual(kl[4],5)
#         kl.append((6,7))
#         self.assertEqual(kl[6],7)
#         kl[2] = 8
#         self.assertEqual(kl[2],8)
#         self.assertEqual(kl.has_key(5),False)
#         self.assertTrue(kl.has_key(6))
#         self.assertEqual(len(kl),3)
#         kl.append((3,8))
#         l = list(kl)
        

#     def test_plotgrouptemplate(self):
#         pgt = PlotGroupTemplate()

#         pt1 = plot.PlotTemplate({'Strength'   : None,
#                                  'Hue'        : 'HueP',
#                                  'Confidence' : None})
#         pt2 = plot.PlotTemplate({'Strength'   : 'HueSel',
#                                  'Hue'        : 'HueP',
#                                  'Confidence' : None})
#         pt3 = plot.PlotTemplate({'Strength'   : 'HueSel',
#                                  'Hue'        : None,
#                                  'Confidence' : None})
#         pgt = PlotGroupTemplate([('HuePref', pt1),
#                                  ('HuePrefAndSel', pt2),
#                                  ('HueSelect', pt3)])

#         pgt2 = PlotGroupTemplate( 
#             [('HuePref', PlotTemplate({'Strength'   : None,
#                                        'Hue'        : 'HueP',
#                                        'Confidence' : None})),
#              ('HuePrefAndSel', PlotTemplate({'Strength'   : 'HueSel',  
#                                              'Hue'        : 'HueP',
#                                              'Confidence' : None})),
#              ('HueSelect', PlotTemplate({'Strength'   : 'HueSel',
#                                          'Hue'        : None,
#                                          'Confidence' : None}))])


################## PREVIOUSLY IN TESTPLOTENGINE:

### JC: My new statements.
import unittest
from topo.base.simulation import Simulation
from topo.plotting.plotgroup import TemplatePlotGroup, ConnectionFieldsPlotGroup

import topo.pattern.random
from topo.learningfn.som import CFPLF_HebbianSOM
from topo.base.cf import CFProjection, CFSheet
from topo.responsefn.optimized import CFPRF_DotProduct_opt
from topo.base.patterngenerator import BoundingBox

### JCALERT! This file has to be re-written when the fundamental changes in plot.py
### plotengine.py and plotgroup.py will be finished.
### for the moment, the tests are commented out...

# class TestPlotEngine(unittest.TestCase):

#       def setUp(self):

#           self.sim = Simulation()

#           CFSheet.nominal_density = 10
          
#           V1 = CFSheet(name='V1')
#           V2 = CFSheet(name='V2')
#           V3 = CFSheet(name='V3')

#           CFProjection.weights_generator = topo.pattern.random.UniformRandom(bounds=BoundingBox(points=((-0.1,-0.1),(0.1,0.1))))
#           CFProjection.weights_generator = topo.pattern.random.UniformRandom(bounds=BoundingBox(points=((-0.1,-0.1),(0.1,0.1))))
#           CFProjection.response_fn = CFPRF_DotProduct_opt()
#           CFProjection.learning_fn = CFPLF_HebbianSOM()

#           self.sim.connect(V1,V2,delay=0.5,connection_type=CFProjection,name='V1toV2')
#           self.sim.connect(V3,V2,delay=0.5,connection_type=CFProjection,name='V2toV3')

#           self.pe = PlotEngine(self.sim)

#       def test_make_plot_group(self):

#           ### Activity PlotGroup
#           pgt1 = plotgroup_templates['Activity']

#           self.pe.make_plot_group('Activity',pgt1,'BasicPlotGroup',None)
#           test_plot_group = BasicPlotGroup(self.sim,pgt1,'Activity',None,None)
#           test = self.pe.plot_group_dict.get('Activity',None)

#           ### JCALERT! HOW TO TEST THAT TWO PLOTGROUP ARE THE SAME?
#           #self.assertEqual(test_plot_group,test)
          
#           ### Orientation Preference PlotGroup
#           pgt2 = plotgroup_templates['Orientation Preference']

#           self.pe.make_plot_group('Orientation Preference',pgt2,'BasicPlotGroup',None)
#           test_plot_group = BasicPlotGroup(self.sim,pgt2,'Orientation Preference',None,None)
#           test = self.pe.plot_group_dict.get('Orientation Preference',None)
#           #self.assertEqual(test_plot_group,test)

#           ### UnitWeight PlotGroup
#           pgt3 = plotgroup_templates['Unit Weights']

#           pg_key1=('Weights','V1',0,0)
#           pg_key2=('Weights','V2',0.4,0.4)
#           pg_key3=('Weights','V3',0.1,0.1)
#           self.pe.make_plot_group(pg_key1,pgt3,'UnitWeightsPlotGroup','V1')
#           self.pe.make_plot_group(pg_key2,pgt3,'UnitWeightsPlotGroup','V2')
#           self.pe.make_plot_group(pg_key3,pgt3,'UnitWeightsPlotGroup','V3')
          
#           test_plot_group1 = UnitWeightsPlotGroup(self.sim,pgt3,pg_key1,'V1',None)
#           test_lambda = lambda s: s.name == 'V2'
#           test_plot_group2 = UnitWeightsPlotGroup(self.sim,pgt3,pg_key2,test_lambda,None)
#           test_plot_group3 = UnitWeightsPlotGroup(self.sim,pgt3,pg_key3,'V3',None)


#           test1 = self.pe.plot_group_dict.get(pg_key1,None)
#           #self.assertEqual(test_plot_group1,test1)

#           test2 = self.pe.plot_group_dict.get(pg_key2,None)
#           #self.assertEqual(test_plot_group2,test2)

#           test3 = self.pe.plot_group_dict.get(pg_key3,None)
#           #self.assertEqual(test_plot_group3,test3)
          
#           pgt4 = plotgroup_templates['Projection']



suite = unittest.TestSuite()
suite.addTest(unittest.makeSuite(TestPlotGroup))

if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(suite)



"""
class TemplatePlotGroupPanel
Panel for displaying preference maps and activity plot groups.

$Id$
"""


import copy

from Tkinter import DISABLED, NORMAL
from tkFileDialog import asksaveasfilename

import param
from param import normalize_path
import paramtk as tk

import topo.command.pylabplot
from topo.plotting.plotgroup import TemplatePlotGroup

from plotgrouppanel import SheetPanel


### CEBALERT: additional dynamic info/right-click problems:
#
# If I open an Activity plot and then measure an orientation map,
# the plot only shows Activity, but the dynamic info includes the or
# pref and selectivity.  That makes sense, given that the plot would
# show that if it were refreshed, but it's confusing. 




# CEBALERT: should  be something in plot or wherever, or maybe I don't see how to get this
# info the easy way from it? Not sure I actually wrote this correctly, anyway.
def available_plot_channels(plot):
    """
    Return the channels+names of the channels that have views.
    """
    available_channels = {}
    for name,channel in plot.channels.items():
        if plot.view_dict.has_key(channel):
            available_channels[name]=channel
    return available_channels




### CB: I'm working here at the moment.

class TemplatePlotGroupPanel(SheetPanel):

    plotgroup_type = TemplatePlotGroup

    strength_only = param.Boolean(default=False,doc="""If true, disables all but the Strength channel of each plot,
disabling all color coding for Strength/Hue/Confidence plots.""")

    ####################################################################
    # CEBALERT: Ugly hack!  Basic idea for method (deleting conf & hue
    # from plot_templates) copied from previous tkgui.  Instead, we
    # probably want a mechanism for doing this from the command line,
    # probably a global 'monochrome' parameter that all plots respect,
    # and then people can modify the plot templates directly if they
    # need more control than that.
    def _strength_only_hack(self):
        # same as superclass, except that if strength only is true
        # it temporarily removes hue&conf from plot_templates
        original_templates = self.plotgroup.plot_templates
        self.plotgroup.plot_templates = copy.deepcopy(self.plotgroup.plot_templates)
        if self.strength_only:
            for name,template in self.plotgroup.plot_templates:
                for c in ['Confidence','Hue']:
                    if c in template:
                        del template[c]
        return original_templates
        
    def refresh_plots(self):
        # same as superclass, except that if strength only is true
        # it temporarily removes hue&conf from plot_templates
        original_templates = self._strength_only_hack()
        super(TemplatePlotGroupPanel,self).refresh_plots()
        self.plotgroup.plot_templates = original_templates
        
    def redraw_plots(self):
        # same as superclass, except that if strength only is true
        # it temporarily removes hue&conf from plot_templates
        original_templates = self._strength_only_hack()
        super(TemplatePlotGroupPanel,self).redraw_plots()
        self.plotgroup.plot_templates = original_templates
            

    ## CB: update init args now we have no pgts.
    def __init__(self,master,plotgroup,**params):

        super(TemplatePlotGroupPanel,self).__init__(master,plotgroup,**params)

        self.pack_param('strength_only',parent=self.control_frame_1,
                        on_set=self.redraw_plots,side='right')
        
        # Display any plots that can be done with existing data, but
        # don't regenerate the SheetViews unless requested
        if self.plotgroup.plot_immediately:
            self.refresh_plots()
        else:
            self.redraw_plots()
            self.display_labels() # should this be called for any redraw? genuinely needs to be
            # called here because labels might never have been drawn.


        #################### RIGHT-CLICK MENU STUFF ####################
        self._sheet_menu.add_command(label="Save image as PNG",
                                     command=self.__save_to_png)

        
        self._sheet_menu.add_command(label="Save image as EPS",
                                     command=self.__save_to_postscript)

        
        self._unit_menu.add_command(label="Print info",
                                    command=self.__print_info)

        # JABALERT: Shouldn't be assuming SHC here; should also work
        # for RGB (or any other channels).
        channel_menus={}
        for chan in ['Strength','Hue','Confidence']:
            newmenu = tk.Menu(self._canvas_menu, tearoff=0)
            self._canvas_menu.add_cascade(menu=newmenu,label=chan+' channel', indexname=chan)

            # The c=chan construction is required so that each lambda has its own copy of the string
            newmenu.add_command(label="Print matrix",          command=lambda c=chan: self.__print_matrix(c))
            newmenu.add_command(label="Plot in sheet coords",  command=lambda c=chan: self.__plot_sheet_matrix(c))
            newmenu.add_command(label="Plot in matrix coords", command=lambda c=chan: self.__plot_matrix(c))
            newmenu.add_command(label="Plot as 3D wireframe",  command=lambda c=chan: self.__plot_matrix3d(c))
            newmenu.add_command(label="Fourier transform",     command=lambda c=chan: self.__fft(c))
            newmenu.add_command(label="Autocorrelation",       command=lambda c=chan: self.__autocorrelate(c))
            newmenu.add_command(label="Histogram",             command=lambda c=chan: self.__histogram(c))
            newmenu.add_command(label="Gradient",              command=lambda c=chan: self.__gradient(c))
            channel_menus[chan]=newmenu

        
        #self._sheet_menu.add_command(label="Print matrix values",
        #                             command=self.__print_matrix)
        #################################################################





    ### JABALERT: Should remove the assumption that the plot will be
    ### SHC (could e.g.  be RGB).
    ### 
    ### Should add the ability to turn off any of the channels
    ### independently (just as the Strength-only button does), and
    ### eventually should allow the user to type in the name of any
    ### SheetView to change the template as desired to visualize any
    ### quantity.
    ###
    # CB: something about these methods does not seem to fit the PlotGroup hierarchy.
    def _canvas_right_click(self,event_info):
        """
        Make whichever of the SHC channels is present in the plot available on the menu.
        """
        super(TemplatePlotGroupPanel,self)._canvas_right_click(event_info,show_menu=False)
        
        if 'plot' in event_info:
            plot = event_info['plot']
            
            available_channels =available_plot_channels(plot) 

            for channel in ('Strength','Hue','Confidence'):
                if channel in available_channels:
                    self._canvas_menu.entryconfig(channel,
                                                  label="%s channel: %s" %
                                                  (channel,str(plot.channels[channel])),state=NORMAL)
                else:
                    self._canvas_menu.entryconfig(channel,
                                                  label="%s channel: None" %
                                                  (channel),state=DISABLED)

            self._canvas_menu.tk_popup(event_info['event'].x_root,
                                       event_info['event'].y_root)


    def __save_to_png(self):
        plot   = self._right_click_info['plot']
        filename = self.plotgroup.filesaver.filename(plot.label(),file_format="png")
        PNG_FILETYPES = [('PNG images','*.png'),('All files','*')]
        snapshot_name = asksaveasfilename(filetypes=PNG_FILETYPES,
                                          initialdir=normalize_path(),
                                          initialfile=filename)

        if snapshot_name:
            plot.bitmap.image.save(snapshot_name)

    # based on routine in editor.py
    def __save_to_postscript(self):
        plot   = self._right_click_info['plot']
        canvas = self._right_click_info['event'].widget
        filename = self.plotgroup.filesaver.filename(plot.label(),file_format="eps")
        POSTSCRIPT_FILETYPES = [('Encapsulated PostScript images','*.eps'),
                                ('PostScript images','*.ps'),('All files','*')]
        snapshot_name = asksaveasfilename(filetypes=POSTSCRIPT_FILETYPES,
                                          initialdir=normalize_path(),
                                          initialfile=filename)

        if snapshot_name:
            canvas.postscript(file=snapshot_name)

    # CB: these methods assume channel has a view (the menu only displays those that do)
    def __fft(self,channel):
        plot = self._right_click_info['plot']
        description = "%s %s at time %s" % (plot.plot_src_name, plot.name, topo.sim.timestr())
        m=plot._get_matrix(channel)
        topo.command.pylabplot.fftplot(m, title="FFT Plot: " + description)

    def __autocorrelate(self,channel):
        plot = self._right_click_info['plot']
        description = "%s %s at time %s" % (plot.plot_src_name, plot.name, topo.sim.timestr())
        m=plot._get_matrix(channel)
        topo.command.pylabplot.autocorrelationplot(m, title="Autocorrelation: " + description)

    def __histogram(self,channel):
        plot = self._right_click_info['plot']
        description = "%s %s at time %s" % (plot.plot_src_name, plot.name, topo.sim.timestr())
        m=plot._get_matrix(channel)
        topo.command.pylabplot.histogramplot(m,title="Histogram: "+ description)

    def __gradient(self,channel):
        plot = self._right_click_info['plot']
        description = "%s %s at time %s" % (plot.plot_src_name, plot.name, topo.sim.timestr())
        m=plot._get_matrix(channel)
        view = plot.view_dict[plot.channels[channel]]
        topo.command.pylabplot.gradientplot(m,title="Gradient: " + description,
                                              cyclic=view.cyclic,cyclic_range=view.cyclic_range)

    def __print_matrix(self,channel):
        plot = self._right_click_info['plot']
        description = "%s %s at time %s" % (plot.plot_src_name, plot.name, topo.sim.timestr())
        print ("#" + description)
        m=plot._get_matrix(channel)
        print m

    def __plot_matrix(self,channel):
        plot = self._right_click_info['plot']
        description = "%s %s at time %s" % (plot.plot_src_name, plot.name, topo.sim.timestr())
        m=plot._get_matrix(channel)
        topo.command.pylabplot.matrixplot(m, title=description)

    def __plot_sheet_matrix(self,channel):
        plot = self._right_click_info['plot']
        description = "%s %s at time %s" % (plot.plot_src_name, plot.name, topo.sim.timestr())
        m=plot._get_matrix(channel)
        topo.command.pylabplot.activityplot(topo.sim[plot.plot_src_name], m, title=description)

    def __plot_matrix3d(self,channel):
        plot = self._right_click_info['plot']
        description = "%s %s at time %s" % (plot.plot_src_name, plot.name, topo.sim.timestr())
        m=plot._get_matrix(channel)
        topo.command.pylabplot.matrixplot3d(m, title=description)

    # CEBALERT: decide if and how to allow any of these functions to be used for getting as many
    # channels' info as possible.
    # e.g. in this one...
    def __print_info(self,channel=None):
        plot = self._right_click_info['plot']
        (r,c),(x,y) = self._right_click_info['coords']
        description ="%s %s, row %d, col %d at time %s: " % (plot.plot_src_name, plot.name, r, c, topo.sim.timestr())

        channels_info = ""
        if channel is None:
            for channel,name in available_plot_channels(plot).items():
                m=plot._get_matrix(channel)
                channels_info+="%s:%f"%(name,m[r,c])
        else:
            m=plot._get_matrix(channel)
            channels_info+="%s:%f"%(plot.channels[channel],m[r,c])
            
        print "%s %s" % (description, channels_info)





    def _dynamic_info_string(self,event_info,basic_text):
        """
        Also print whatever other channels are there and have views.
        """
        plot = event_info['plot']
        r,c = event_info['coords'][0]
        
        info_string = basic_text

        for channel,channel_name in available_plot_channels(plot).items():
            info_string+=" %s: % 1.3f"%(channel_name,plot._get_matrix(channel)[r,c])
                                        
        return info_string

    






"""

$Id$
"""

# CEBALERT: incomplete!

import unittest
from numpy.testing import assert_array_almost_equal

from topo import pattern 
import topo.pattern.audio

class TestAudio(unittest.TestCase):

    def setUp(self):
        self.audio = pattern.audio.AudioFile(filename="sounds/complex/daisy.wav")

    def test_basic(self):
        result = self.audio()


suite = unittest.TestSuite()
suite.addTest(unittest.makeSuite(TestAudio))



"""
Contains tests to check a particular script's results or speed have
not changed.

Buildbot shows how to run all these tests.

$Id$
"""


import pickle, copy, __main__, timeit, os, os.path, socket, cPickle, inspect, traceback, tempfile, shutil

from numpy.testing import assert_array_equal, assert_array_almost_equal

import param
from param import resolve_path, normalize_path

import topo


# CEBALERT: should get this from somewhere else!
TOPOGRAPHICAHOME = os.path.join(os.path.expanduser("~"),"topographica")

TESTSDATADIR = os.path.join(TOPOGRAPHICAHOME,"tests")
if not os.path.exists(TESTSDATADIR):
    os.makedirs(TESTSDATADIR)

# While training data is usually checked into topo/tests and is the
# same for all machines, speed data is generated by the machine
# running this makefile. Therefore, speed data is stored in a
# machine-specific directory.

MACHINETESTSDATADIR = os.path.join(TESTSDATADIR,socket.gethostname())
if not os.path.exists(MACHINETESTSDATADIR):
    os.makedirs(MACHINETESTSDATADIR)

FIXEDDATADIR = resolve_path("topo/tests",path_to_file=False)



######################################################################################
### Support fns

def _support_old_args(args):
    # support old data files which contain 'default_density', etc
    if 'default_density' in args:
        args['cortex_density']=args['default_density']
        #del args['default_density']
    if 'default_retina_density' in args:
        args['retina_density']=args['default_retina_density']
        #del args['default_retina_density']
    if 'default_lgn_density' in args:
        args['lgn_density']=args['default_lgn_density']
        #del args['default_lgn_density']
    # (left the dels commented out for now in case scripts still use old names)

def _setargs(args):
    for arg,val in args.items():
        print "Setting %s=%s"%(arg,val)
        __main__.__dict__[arg]=val

# For generating data in a separate process, leaving the parent with
# the original settings (i.e. avoids having to do a reset of
# simulation time, etc).
# CEBALERT: is this somehow causing func to run more slowly than
# without forking?
def _run_in_forked_process(func, *args, **kwds):
    # DSALERT: os.fork() is not supported on Windows
    pid = os.fork()
    if pid > 0:
        os.waitpid(pid, 0)
    else: 
        func(*args, **kwds)
        os._exit(0)


def _instantiate_everything(
    classes_to_exclude=("topo.base.simulation.Simulation","topo.base.simulation.Simulation"),
    modules_to_exclude=('plotting','tests','tkgui','command','util')):

    # default excludes currently set up for pickle tests

    # CEBALERT: this is basically get_PO_class_attributes from param.parameterized
    def get_classes(module,classes,processed_modules,module_excludes=()):
        exec "from %s import *"%module.__name__ in locals()
        dict_ = module.__dict__
        for (k,v) in dict_.items():
            if '__all__' in dict_ and inspect.ismodule(v) and k not in module_excludes:
                if k in dict_['__all__'] and v not in processed_modules:
                    get_classes(v,classes,processed_modules,module_excludes)
                processed_modules.append(v)
            else:
                # class & not parameterizedfunction & not __abstract & not excluded & starts with topo. or param.
                if isinstance(v,type) and not isinstance(v,param.ParameterizedFunction) and not (hasattr(v,"_%s__abstract"%v.__name__) and getattr(v,"_%s__abstract"%v.__name__) is True):
                    full_class_path = v.__module__+'.'+v.__name__
                    if (not full_class_path in classes) and (not full_class_path in classes_to_exclude) and (full_class_path.startswith("topo") or full_class_path.startswith("param")):
                        classes.append(full_class_path)

    classes = []
    processed_modules = []
     
    import topo
    get_classes(topo,classes,processed_modules,module_excludes=modules_to_exclude)
    get_classes(param,classes,processed_modules,module_excludes=modules_to_exclude)

    instances = []

    instantiated_names = []
    uninstantiated_names = []
     
    for class_name in classes:
        try:
            instances.append(eval(class_name+"()"))
            instantiated_names.append(class_name)
        except:
            #print "Could not instantiate %s"%class_name
            uninstantiated_names.append(class_name)

    print "\n ** Instantiated %s classes:"%len(instantiated_names)
    print "\n".join(instantiated_names)

    print "\n ** Could not instantiate %s classes:"%len(uninstantiated_names)
    print "\n".join(uninstantiated_names)
    
    return instances

######################################################################################


# CEBALERT: document somewhere about when to delete data files
# (i.e. when to generate new data) for train-tests and speed-tests and
# startup-speed-tests.


######################################################################################
### train-tests

RUN_FOR = [1,99,150]
LOOK_AT = "V1"
TRAINTESTS_CORTEXDENSITY = 8
RETINA_DENSITY = 24
LGN_DENSITY = 24

def _generate_data(script,data_filename,look_at='V1',run_for=[1,99,150],**args):
    
    print "Generating data for %s's %s after topo.sim.run(%s)"%(script,look_at,run_for)
    
    _setargs(args)

    execfile(script,__main__.__dict__)
    
    data = {}

    for time in run_for:
        print "Running for %s iterations"%time
        topo.sim.run(time)
        print "Recording data for %s at %s"%(look_at,topo.sim.timestr())
        data[topo.sim.timestr()] = copy.deepcopy(topo.sim[look_at].activity)

    data['args']=args
    data['run_for']=run_for
    data['look_at']=look_at
    data['versions'] = topo.version,topo.release

    print "Saving data to %s"%data_filename
    pickle.dump(data,open(data_filename,'wb'),2)



def test_script(script,decimal=None):
    """
    Run script with the parameters specified when its DATA file was
    generated, and check for changes.

    Looks for the DATA file at FIXEDDATADIR/script_name.ty_DATA (for
    data checked into SVN). If not found there, looks at
    TESTSDATADIR/script_name.ty_DATA. If also not found there, first
    generates a new DATA file at TESTSDATADIR/script_name.ty_DATA
    (i.e. to generate new data, delete the existing data before running).

    The decimal parameter defines how many decimal points to use when
    testing for array equality. The default of None causes exact
    matching.
    """
    print "Comparing results for %s"%script
    
    script_name = os.path.basename(script)
    # CEBALERT: clean up
    data_filename_only = script_name+"_DATA"
    data_filename = os.path.join(TESTSDATADIR,data_filename_only)

    try:
        locn = resolve_path(data_filename_only,search_paths=[FIXEDDATADIR,TESTSDATADIR])
    except IOError:
        print "No existing data"
        _run_in_forked_process(_generate_data,script,data_filename,run_for=RUN_FOR,cortex_density=TRAINTESTS_CORTEXDENSITY,lgn_density=LGN_DENSITY,retina_density=RETINA_DENSITY)
        locn = resolve_path(data_filename)
            
    print "Reading data from %s"%locn

    data_file = open(locn,'rb')
    data = pickle.load(data_file)

    print "Data from release=%s, version=%s"%(data['versions'] if 'versions' in data else ("unknown","unknown"))

    # retrieve parameters used when script was run
    run_for=data['run_for']
    look_at = data['look_at']

    ####################################################
    # support very old data files that contain 'density' instead of args['cortex_density']
    if 'args' not in data:
        data['args']={'cortex_density' : data['density']}

    args = data['args']
    _support_old_args(args)
    ####################################################

    _setargs(args)

    print "Starting '%s'"%script
    execfile(script,__main__.__dict__)        

    #########################################################
    time_fmt = topo.sim.timestr
    # support old pickled data (could replace time_fmt(topo.sim.time()) with
    # just topo.sim.timestr() if we didn't need to support old data
    if topo.sim.timestr(run_for[0]) not in data:
        time_fmt = float
    #########################################################

    for time in run_for:
        print "Running for %s iterations"%time
        topo.sim.run(time)
        

        if decimal is None:
            assert_array_equal(data[time_fmt(topo.sim.time())],topo.sim[look_at].activity,
                           err_msg="\nAt topo.sim.time()=%d, with decimal=%s"%(topo.sim.time(),decimal))
        else:
            assert_array_almost_equal(data[time_fmt(topo.sim.time())],topo.sim[look_at].activity,
                           decimal,err_msg="\nAt topo.sim.time()=%d, with decimal=%s"%(topo.sim.time(),decimal))

    result = "Results from " + script + " have not changed."
    if decimal is not None: result+= " (%d dp)" % (decimal)
    print result+"\n"


# CEBALERT: old name
#TestScript = test_script


###########################################################################
### speed-tests

SPEEDTESTS_CORTEXDENSITY=48
SPEEDTESTS_ITERATIONS = 250

# CEBALERT: see ALERT about variation by time_sim_startup()
def _time_sim_run(script,iterations=10):
    """
    Execute the script in __main__, then time topo.sim.run(iterations).

    Uses the timeit module.
    """
    print "Running '%s' for %s iterations"%(script,iterations)
    execfile(script,__main__.__dict__)
    topo.sim.run(1) # ensure compilations etc happen outside timing
    # CB: we enable garbage collection
    # (http://docs.python.org/lib/module-timeit.html)
    return timeit.Timer('topo.sim.run('+`iterations`+')','gc.enable(); import topo').timeit(number=1)


def _generate_speed_data(script,data_filename,iterations=100,**args):
    print "Generating speed data for %s"%script
    
    _setargs(args)
                
    how_long = _time_sim_run(script,iterations)

    speed_data = {'args':args,
                  'iterations':iterations,
                  'how_long':how_long}

    speed_data['versions'] = topo.version,topo.release
                          
    print "Saving data to %s"%data_filename
    pickle.dump(speed_data,open(data_filename,'wb'),2)


def compare_speed_data(script):
    """
    Run and time script with the parameters specified when its SPEEDDATA file was
    generated, and check for changes.

    Looks for the SPEEDDATA file at
    MACHINETESTSDATADIR/script_name.ty_DATA. If not found there, first
    generates a new SPEEDDATA file at
    MACHINETESTSDATADIR/script_name.ty_DATA (i.e. to generate new
    data, delete the existing data before running).
    """
    print "Comparing speed data for %s"%script
    
    script_name = os.path.basename(script)
    data_filename = os.path.join(MACHINETESTSDATADIR,script_name+"_SPEEDDATA")

    try:
        locn = resolve_path(data_filename)
    except IOError:
        print "No existing data"
        _run_in_forked_process(_generate_speed_data,script,data_filename,iterations=SPEEDTESTS_ITERATIONS,cortex_density=SPEEDTESTS_CORTEXDENSITY)
        #_generate_speed_data(script,data_filename,iterations=SPEEDTESTS_ITERATIONS,cortex_density=SPEEDTESTS_CORTEXDENSITY)
        locn = resolve_path(data_filename)
            
    print "Reading data from %s"%locn

    speed_data_file = open(locn,'r')

    try:
        speed_data = pickle.load(speed_data_file)
        print "Data from release=%s, version=%s"%(speed_data['versions'] if 'versions' in speed_data else ("unknown","unknown"))
    except:
    ###############################################################
    ## Support old data files (used to be string in the file rather
    ## than pickle)
        speed_data_file.seek(0)
        speed_data = speed_data_file.readline()

        iterations,old_time = speed_data.split('=')
        iterations = float(iterations); old_time=float(old_time)
        speed_data = {'iterations':iterations,
                      'how_long':old_time,
                      'args':{}}
    ###############################################################
        
    speed_data_file.close()
        
    old_time = speed_data['how_long']
    iterations = speed_data['iterations']
    args = speed_data['args']    

    _support_old_args(args)

    _setargs(args)

    new_time = _time_sim_run(script,iterations)

    percent_change = 100.0*(new_time-old_time)/old_time

    print "["+script+"]"+ '  Before: %2.1f s  Now: %2.1f s  (change=%2.1f s, %2.1f percent)'\
          %(old_time,new_time,new_time-old_time,percent_change)

    # CEBALERT: whatever compensations the python timing functions are supposed to make for CPU
    # activity, do they work well enough? If the processor is being used, these times jump all
    # over the place (i.e. vary by more than 10%).
    #assert percent_change<=5, "\nTime increase was greater than 5%"

###########################################################################


###########################################################################
### startup timing

# CEBALERT: figure out what this meant: "expect variation in these
# results! see python's timeit module documentation"
def _time_sim_startup(script):
    print "Starting %s"%script
    return timeit.Timer("execfile('%s',__main__.__dict__)"%script,'import __main__;gc.enable()').timeit(number=1)

     
def _generate_startup_speed_data(script,data_filename,**args):
    print "Generating startup speed data for %s"%script

    _setargs(args)
    how_long = _time_sim_startup(script)

    speed_data = {'args':args,
                  'how_long':how_long}

    speed_data['versions'] = topo.version,topo.release

    print "Saving data to %s"%data_filename
    pickle.dump(speed_data,open(data_filename,'wb'),2)


def compare_startup_speed_data(script):
    """
    Run and time script with the parameters specified when its
    STARTUPSPEEDDATA file was generated, and check for changes.

    Looks for the STARTUPSPEEDDATA file at
    MACHINETESTSDATADIR/script_name.ty_STARTUPSPEEDDATA. If not found
    there, first generates a new STARTUPSPEEDDATA file at
    MACHINETESTSDATADIR/script_name.ty_STARTUPSPEEDDATA (i.e. to
    generate new data, delete the existing data before running).
    """
    print "Comparing startup speed data for %s"%script
    
    script_name = os.path.basename(script)
    data_filename = os.path.join(MACHINETESTSDATADIR,script_name+"_STARTUPSPEEDDATA")

    try:
        locn = resolve_path(data_filename)
    except IOError:
        print "No existing data"
        _run_in_forked_process(_generate_startup_speed_data,script,data_filename,cortex_density=SPEEDTESTS_CORTEXDENSITY)        
        #_generate_startup_speed_data(script,data_filename,cortex_density=SPEEDTESTS_CORTEXDENSITY)
        locn = resolve_path(data_filename)
            
    print "Reading data from %s"%locn

    speed_data_file = open(locn,'r')

    try:
        speed_data = pickle.load(speed_data_file)
        print "Data from release=%s, version=%s"%(speed_data['versions'] if 'versions' in speed_data else ("unknown","unknown"))
    except:
    ###############################################################
    ## Support old data files (used to be string in the file rather
    ## than pickle)
        speed_data_file.seek(0)
        speed_data = speed_data_file.readline()

        density,old_time = speed_data.split('=')
        speed_data = {'cortex_density':float(density),
                      'how_long':float(old_time),
                      'args':{}}

    _support_old_args(speed_data['args'])
    ###############################################################

    _setargs(speed_data['args'])

    speed_data_file.close()

    old_time = speed_data['how_long']        
    new_time = _time_sim_startup(script)

    percent_change = 100.0*(new_time-old_time)/old_time

    print "["+script+ ' startup]  Before: %2.1f s  Now: %2.1f s  (change=%2.1f s, %2.1f percent)'\
          %(old_time,new_time,new_time-old_time,percent_change)


### end startup timing
###########################################################################


###########################################################################
### Snapshot tests

# This is clumsy. We could control topographica subprocesses, but I
# can't remember how to do it

def compare_with_and_without_snapshot_NoSnapshot(script="models/lissom.ty",look_at='V1',cortex_density=8,lgn_density=4,retina_density=4,dims=['or','od','dr','cr','dy','sf'],dataset="Gaussian",run_for=10,break_at=5):

    data_filename=os.path.split(script)[1]+"_PICKLETEST"
    
    # we must execute in main because e.g. scheduled events are run in __main__
    # CEBALERT: should set global params
    __main__.__dict__['cortex_density']=cortex_density
    __main__.__dict__['lgn_density']=lgn_density
    __main__.__dict__['retina_density']=retina_density
    __main__.__dict__['dims']=dims
    __main__.__dict__['dataset']=dataset
    
    execfile(script,__main__.__dict__)
    
    data = {}
    topo.sim.run(break_at)
    data[topo.sim.time()]= copy.deepcopy(topo.sim[look_at].activity)
    topo.sim.run(run_for-break_at)
    data[topo.sim.time()]= copy.deepcopy(topo.sim[look_at].activity)
        
    data['run_for']=run_for
    data['break_at']=break_at
    data['look_at']=look_at

    data['cortex_density']=cortex_density
    data['lgn_density']=lgn_density
    data['retina_density']=retina_density
    data['dims']=dims
    data['dataset']=dataset
    
    locn = normalize_path(os.path.join("tests",data_filename))
    print "Writing pickle to %s"%locn
    pickle.dump(data,open(locn,'wb'),2)


def compare_with_and_without_snapshot_CreateSnapshot(script="models/lissom.ty"):
    data_filename=os.path.split(script)[1]+"_PICKLETEST"

    locn = resolve_path(os.path.join('tests',data_filename))
    print "Loading pickle at %s"%locn
        
    try:
        data = pickle.load(open(locn,"rb"))
    except IOError:
        print "\nData file '"+data_filename+"' could not be opened; run _A() first."
        raise

    # retrieve parameters used when script was run
    run_for=data['run_for']
    break_at=data['break_at']
    look_at=data['look_at']

    # CEBALERT: shouldn't need to re-list - should be able to read from data!
    cortex_density=data['cortex_density']
    lgn_density=data['lgn_density']
    retina_density=data['retina_density']
    dims=data['dims']
    dataset=data['dataset']

    __main__.__dict__['cortex_density']=cortex_density
    __main__.__dict__['lgn_density']=lgn_density
    __main__.__dict__['retina_density']=retina_density
    __main__.__dict__['dims']=dims
    __main__.__dict__['dataset']=dataset
    execfile(script,__main__.__dict__)        

    # check we have the same before any pickling
    topo.sim.run(break_at)
    assert_array_equal(data[topo.sim.time()],topo.sim[look_at].activity,
                       err_msg="\nAt topo.sim.time()=%d"%topo.sim.time())

    from topo.command import save_snapshot
    locn = normalize_path(os.path.join('tests',data_filename+'.typ_'))
    print "Saving snapshot to %s"%locn
    save_snapshot(locn)


def compare_with_and_without_snapshot_LoadSnapshot(script="models/lissom.ty"):
    data_filename=os.path.split(script)[1]+"_PICKLETEST"
    snapshot_filename=os.path.split(script)[1]+"_PICKLETEST.typ_"

    locn = resolve_path(os.path.join('tests',data_filename))
    print "Loading pickle from %s"%locn
    try:
        data = pickle.load(open(locn,"rb"))
    except IOError:
        print "\nData file '"+data_filename+"' could not be opened; run _A() first"
        raise

    # retrieve parameters used when script was run
    run_for=data['run_for']
    break_at=data['break_at']
    look_at=data['look_at']

#    # CEBALERT: shouldn't need to re-list - should be able to read from data!
#    cortex_density=data['cortex_density']
#    lgn_density=data['lgn_density']
#    retina_density=data['retina_density']
#    dims=data['dims']
#    dataset=data['dataset']
    
    from topo.command import load_snapshot

    locn = resolve_path(os.path.join('tests',snapshot_filename))
    print "Loading snapshot at %s"%locn

    try:
        load_snapshot(locn)
    except IOError:
        print "\nPickle file '"+snapshot_filename+"' could not be opened; run _B() first."
        raise

    assert topo.sim.time()==break_at
    assert_array_equal(data[topo.sim.time()],topo.sim[look_at].activity,
                       err_msg="\nAt topo.sim.time()=%d"%topo.sim.time())
    print "Match at %s after loading snapshot"%topo.sim.time()

    topo.sim.run(run_for-break_at)
                
    assert_array_equal(data[topo.sim.time()],topo.sim[look_at].activity,
                       err_msg="\nAt topo.sim.time()=%d"%topo.sim.time())

    print "Match at %s after running loaded snapshot"%topo.sim.time()

### end Snapshot tests
###########################################################################


###########################################################################
### pickle tests
def pickle_unpickle_everything(existing_pickles=None):

    pickle_errors = 0

    if existing_pickles is None:

        instances = _instantiate_everything()
        pickles = {}

        for instance in instances:
            try:
                pickles[str(instance)]=pickle.dumps(instance)
            except:
                print "Error pickling %s:"%instance
                pickle_errors+=1
                traceback.print_exc()
    else:
        pickles = pickle.load(open(existing_pickles))

    unpickle_errors = 0
    
    for instance_name,pickled_instance in pickles.items():
        try:
            pickle.loads(pickled_instance)
        except:
            print "Error unpickling %s"%instance_name
            unpickle_errors+=1
            traceback.print_exc()

    print

    if existing_pickles is None:
        print "Instances that failed to pickle: %s"%pickle_errors
        
    print "Pickled instances that failed to unpickle: %s"%unpickle_errors

    return pickle_errors+unpickle_errors
###########################################################################



###########################################################################
# basic test of run batch
def test_runbatch():
    from topo.misc.genexamples import find_examples
    from topo.command import run_batch

    original_output_path = param.normalize_path.prefix
    start_output_path = tempfile.mkdtemp()
    param.normalize_path.prefix = start_output_path
    
    tiny = os.path.join(find_examples(),"tiny.ty")
    run_batch(tiny,cortex_density=1,retina_density=1,times=[1],snapshot=True,output_directory="testing123")

    new_output_path = param.normalize_path.prefix

    assert new_output_path.startswith(start_output_path)
    assert "testing123" in new_output_path # not perfect test, but better than nothing.

    base = os.path.basename(new_output_path).split(",")[0]

    def exists(endpart):
        whole = os.path.join(new_output_path,base+endpart)
        print "Checking for %s"%whole
        return os.path.isfile(whole)
    
    assert exists(".global_params.pickle")
    assert exists(".out")
    assert exists("_000001.00_V1_Activity.png")
    assert exists("_000001.00_script_repr.ty")
    assert exists("_000001.00.typ")

    print "Deleting %s"%param.normalize_path.prefix
    shutil.rmtree(param.normalize_path.prefix)
    param.normalize_path.prefix=original_output_path
###########################################################################





###########################################################################
## CEBALERT: for C++ reference simulations - should be moved elsewhere
def run_multiple_density_comparisons(ref_script):
    from topo.misc.util import cross_product
    import subprocess
    import traceback
    import os
    
    #k = [8,10,12,13,14,34]
    #x = cross_product([k,k])

    
    x = [[ 8, 8],[ 8, 9],[ 8,10],[ 8,11],[ 8,12],[ 8,13],[ 8,14],[ 8,15],
         [24,14],[24,17],[24,20],
         [24,24],[24,48]]

    cmds = []
    for spec in x:
        c="""./topographica -c "verbose=False;BaseRN=%s;BaseN=%s;comparisons=True;stop_at_1000=False" topo/tests/reference/%s"""%(spec[0],spec[1],ref_script)
        cmds.append(c)

    results = []
    errs=[]
    for cmd in cmds:
        print 
        print "************************************************************"
        print "Executing '%s'"%cmd
        
#        errout = os.tmpfile()#StringIO.StringIO()

        p = subprocess.Popen(cmd, shell=True,stderr=subprocess.PIPE)
        p.wait()
        r = p.returncode
        
        errout = p.stderr
        #r = subprocess.call(cmd,shell=True)#,stderr=subprocess.PIPE)#errout)
        #print "TB",traceback.print_exc()
        
        if r==0:
            result = "PASS"
        else:
            result = "FAIL"
        results.append(result)

        l = errout.readlines()
        i = 0
        L=0
        for line in l:
            if line.startswith("AssertionError"):
                L=i
                break
            i+=1
        
        errs.append(l[L::])
        errout.close()

    print "================================================================================"
    print
    print "SUMMARY"
    print
    nerr = 0
    for xi,result,err in zip(x,results,errs):
        print
        print "* %s ... BaseRN=%s,BaseN=%s"%(result,xi[0],xi[1])
        if result=="FAIL":
            e = ""
            print e.join(err)
            
            nerr+=1
    print "================================================================================"
    
    return nerr

###########################################################################

